Package: BBNI
Title: Bayesian Inference of Boolean Genetic Networks
Version: 0.2.2
Authors@R: c(
    person(given = "Anson", family = "Li",
           email = "liyuanrui618@gmail.com",
           role = c("aut", "cre")),
    person(given = "Shengtong", family = "Han",
           email = "shengtong.han@marquette.edu",
           role = "aut")
    )
Description: Implements a fully Bayesian Markov chain Monte Carlo (MCMC) approach for inferring the topology and Boolean logic transition functions of gene regulatory networks from noisy, binary time-series expression data. Network structure and Boolean rules are sampled jointly from their posterior distribution, providing principled uncertainty quantification rather than a single point estimate. Method described in Han et al. (2014) <doi:10.1371/journal.pone.0115806>.
Depends: R (>= 4.1.0)
License: BSD_3_clause + file LICENSE
URL: https://anson-li8.github.io/BBNI/,
        https://github.com/anson-li8/BBNI
BugReports: https://github.com/anson-li8/BBNI/issues
Encoding: UTF-8
Language: en-US
Imports: stats, igraph
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
LazyData: true
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-08-17 16:53:18 UTC; Anson
Author: Anson Li [aut, cre],
  Shengtong Han [aut]
Maintainer: Anson Li <liyuanrui618@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-19 19:20:12 UTC
Built: R 4.6.1; ; 2026-08-19 23:15:55 UTC; unix

# Generated by roxygen2: do not edit by hand

export(GenerateNetwork)
export(GenerateSample)
export(plot_bbni)
export(plot_network)
export(plot_trace)
export(run_bbni)
importFrom(graphics,
  abline,
  legend,
  plot
)
importFrom(igraph,
  graph_from_adjacency_matrix,
  plot.igraph
)
importFrom(stats,
  rbinom,
  runif
)
importFrom(utils,
  setTxtProgressBar,
  txtProgressBar
)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# BBNI

<!-- badges: start -->

[![R-CMD-check](https://github.com/anson-li8/BBNI/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/anson-li8/BBNI/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/BBNI)](https://CRAN.R-project.org/package=BBNI)
<!-- badges: end -->

BBNI (<u>B</u>ayesian <u>B</u>oolean <u>N</u>etwork <u>I</u>nference) is
a Bayesian approach to Boolean gene regulatory network inference from
noisy gene expression data. The method is discussed in more detail in
[Han et al. (2014)](https://doi.org/10.1371/journal.pone.0115806).
Unlike methods that return a single best-fit network topology, such as
REVEAL and BFE, found in
[BoolNet](https://cran.r-project.org/package=BoolNet), BBNI uses Markov
chain Monte Carlo (MCMC) to sample from a joint posterior distribution
of network topologies and Boolean transition functions. BBNI
deliberately takes biological noise into account and allows for
summarization (such as Bayesian model averaging, or BMA) that stabilizes
results through posterior edge probabilities rather than a single point
estimate.

## Installation

Install the stable version from CRAN:

``` r
install.packages("BBNI")
```

The development version can be installed from GitHub:

``` r
# install.packages("devtools")
devtools::install_github("anson-li8/BBNI")
```

## Performance

BBNI’s MCMC sampler was significantly optimized after the initial CRAN
release. Controlled benchmarks show roughly a 10-fold speedup for
20-node networks; exact runtimes depend on hardware and system load.

## Example

The following example provides a minimal check that the package loads
successfully and can generate simulated data:

``` r
library(BBNI)

set.seed(1)
true_network <- GenerateNetwork(num.node = 5)
dummy_data <- GenerateSample(
  trans_matrix = true_network,
  num.node = 5,
  SampleSize = 20,
  para = rep(0.5, 5),
  error = matrix(0, nrow = 5, ncol = 20)
)
dummy_data
#>      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10] [,11] [,12] [,13] [,14]
#> [1,]    0    0    0    0    0    1    1    1    0     0     1     1     1     0
#> [2,]    0    0    0    0    0    0    0    1    0     0     0     0     1     0
#> [3,]    0    1    1    1    1    1    0    0    0     1     1     0     0     0
#> [4,]    0    0    0    0    0    1    1    0    0     0     1     1     0     0
#> [5,]    1    1    1    1    0    0    1    1    1     0     0     1     1     0
#>      [,15] [,16] [,17] [,18] [,19] [,20]
#> [1,]     1     1     1     1     0     0
#> [2,]     0     0     0     1     0     0
#> [3,]     1     0     0     0     0     1
#> [4,]     1     1     1     0     0     0
#> [5,]     0     0     1     1     1     1
```

For a complete demonstration running the MCMC sampler and evaluating
final convergence and overall network recovery, see the [Introduction to
BBNI](https://anson-li8.github.io/BBNI/articles/Introduction_to_BBNI.html)
vignette.

## Citation

    To cite package 'BBNI' in publications use:

      Han S, Wong RKW, Lee TCM, Shen L, Li S-YR, Fan X (2014). A Full
      Bayesian Approach for Boolean Genetic Network Inference. PLoS ONE
      9(12): e115806. doi:10.1371/journal.pone.0115806

## License

BSD-3-Clause

#' Plot the Inferred Bayesian Boolean Network
#'
#' @description
#' Visualizes the causal network structure inferred by the BBNI MCMC sampler.
#' This function takes the marginal posterior probability of each directed edge
#' and plots the network using the \code{igraph} package. Edges with posterior
#' probabilities below the specified threshold are omitted from the plot.
#'
#' @param results The list returned by \code{run_bbni()}, containing \code{networks} and \code{log_posterior}.
#' @param threshold Numeric. The minimum posterior probability required to draw an edge. Defaults to 0.5.
#' @param node_names Character vector. Optional names for the nodes. Defaults to "N1", "N2", etc.
#' @param true_network Optional square matrix representing the true network topology. If provided, edges will be color-coded to indicate true positives (along with displaying wrong function inferences), false positives, and false negatives. Purely for simulation purposes.
#' @param ... Additional graphical parameters passed to \code{igraph::plot.igraph()}.
#'
#' @return An invisible \code{igraph} object.
#'
#' @examples
#' # 1. Generate synthetic network and time-series data
#' set.seed(123)
#' true_network <- GenerateNetwork(num.node = 5)
#' dummy_data <- GenerateSample(true_network, SampleSize = 15)
#'
#' # 2. Run BBNI sampler
#' prior_para <- matrix(3, nrow = 6, ncol = 2)
#' prior_para[6, 1] <- 2
#' prior_para[6, 2] <- 100
#'
#' results <- run_bbni(dummy_data, prior_para = prior_para, num_update = 100)
#'
#' # 3. Plot inferred network
#' plot_bbni(results, true_network = true_network, threshold = 0.5)
#'
#' @importFrom igraph graph_from_adjacency_matrix plot.igraph
#' @export
plot_bbni <- function(results, threshold = 0.5, node_names = NULL, true_network = NULL, ...) {
  # Make sure igraph is available
  if (!requireNamespace("igraph", quietly = TRUE)) {
    stop("Package \"igraph\" is required for plotting. Please install it.", call. = FALSE)
  }
  # Extract list of networks and calculate posterior probabilities (res)
  if (!is.null(results$post_edge_prob)) {
    res <- results$post_edge_prob
  } else {
    # fallback by manually calculate
    networks <- results$networks
    res <- Reduce(`+`, lapply(networks, function(m) (m > 0) * 1)) / length(networks)
  }
  # Filter probability matrix based on threshold
  adj_matrix <- ifelse(res >= threshold, 1, 0)
  # Assign node names if provided, otherwise default to N1, N2...
  num_nodes <- nrow(res)
    if (is.null(node_names)) {
    colnames(adj_matrix) <- rownames(adj_matrix) <- paste0("N", 1:num_nodes)
    vsize <- 12   # short N-labels: keep current look
    lcex  <- 0.7
  } else {
    colnames(adj_matrix) <- rownames(adj_matrix) <- node_names
    vsize <- 20   # real gene names (e.g. CDC20), bigger circle + smaller text so labels fit
    lcex  <- 0.6
  }
  # get functions for inferred edges
  inf_func_matrix <- matrix(0, num_nodes, num_nodes)
  # filter out burn-in samples so function inference isn't polluted by early chain states
  total_samples <- length(results$networks)
  burn_in_ratio <- if (!is.null(results$burn_in)) results$burn_in else 0
  burn_in_steps <- floor(burn_in_ratio * (total_samples - 1))
  post_samples <- results$networks[(burn_in_steps + 2):total_samples]
  for (i in 1:num_nodes) {
    for (j in 1:num_nodes) {
      if (adj_matrix[i, j] == 1) {
        vals <- sapply(post_samples, `[` , i, j)
        vals <- vals[vals > 0]
        if (length(vals) > 0) {
          inf_func_matrix[i, j] <- as.integer(names(which.max(table(vals))))
        }
      }
    }
  }
  # if ground truth provided, color-code to show model performance
  if (!is.null(true_network)) {
    true_adj <- (true_network > 0) * 1
    combined_adj <- (adj_matrix | true_adj) * 1
    g <- igraph::graph_from_adjacency_matrix(t(combined_adj), mode = "directed")
    edges <- igraph::as_edgelist(g, names = FALSE)
    idx <- cbind(edges[, 2], edges[, 1])
    is_true <- true_adj[idx] == 1
    is_inf  <- adj_matrix[idx] == 1
    is_correct_func <- (true_network[idx] == inf_func_matrix[idx])
    e_colors <- rep("darkgray", igraph::ecount(g))
    e_colors[is_true & is_inf & is_correct_func] <- "darkgreen"   # True Positive
    e_colors[is_true & is_inf & !is_correct_func] <- "darkorange" # True Positive, but wrong function
    e_colors[!is_true & is_inf] <- "firebrick"  # False Positive
    e_colors[is_true & !is_inf] <- "gray70"     # False Negative (Missed)
    e_lty <- rep(1, igraph::ecount(g))
    e_lty[is_true & !is_inf] <- 2               # Dashed for missed edges
    igraph::E(g)$color <- e_colors
    igraph::E(g)$lty <- e_lty
    main_title <- paste("Inferred vs True Network (Threshold >=", threshold, ")")
  } else {
    # default plotting w/o ground truth
    g <- igraph::graph_from_adjacency_matrix(t(adj_matrix), mode = "directed")
    igraph::E(g)$color <- "darkgray"
    igraph::E(g)$lty <- 1
    main_title <- paste("Inferred Network (Threshold >=", threshold, ")")
  }
  # plot graph
    # Capture extra arguments (like layout)
  dots <- list(...)
  # plot graph
  do.call(igraph::plot.igraph, c(list(g,
    vertex.size = vsize,
    vertex.color = "lightblue",
    vertex.label.color = "black",
    vertex.label.cex = lcex,
    edge.arrow.size = 0.4,
    edge.curved = 0.1,
    main = main_title), dots))
  # add legend if ground truth was provided
  if (!is.null(true_network)) {
    legend("bottomleft",
           legend = c("Correct (TP)", "Correct (TP), Wrong Function", "Spurious (FP)", "Missed (FN)"),
           col = c("darkgreen", "darkorange", "firebrick", "gray70"),
           lty = c(1, 1, 1, 2), lwd = 2, bty = "n", cex = 0.8)
  }
  # return graph object invisibly in case user wants to manipulate it further
  invisible(g)
}

#' Plot a Single Boolean Network
#'
#' Visualizes a single directed network topology. For instance, a synthetic "true" network
#' generated by \code{GenerateNetwork()}.
#'
#' @param trans_matrix A square matrix representing the network topology and Boolean rules (e.g., the output of \code{GenerateNetwork()}). Elements greater than 0 indicate a directed edge.
#' @param node_names Character vector. Optional names for the nodes. Defaults to "N1", "N2", etc.
#' @param ... Additional graphical parameters passed to \code{igraph::plot.igraph()}.
#'
#' @return An invisible \code{igraph} object.
#'
#' @examples
#' # 1. Generate synthetic 5-node network topology
#' set.seed(123)
#' true_network <- GenerateNetwork(num.node = 5)
#'
#' # 2. Plot the generated network
#' plot_network(true_network)
#'
#' @importFrom igraph graph_from_adjacency_matrix plot.igraph
#' @export
plot_network <- function(trans_matrix, node_names = NULL, ...) {
  # Make sure igraph is available
  if (!requireNamespace("igraph", quietly = TRUE)) {
    stop("Package \"igraph\" is required for plotting. Please install it.", call. = FALSE)
  }
  adj_matrix <- (trans_matrix > 0) * 1
  num_nodes <- nrow(adj_matrix)
  if (is.null(node_names)) {
    colnames(adj_matrix) <- rownames(adj_matrix) <- paste0("N", 1:num_nodes)
  } else {
    colnames(adj_matrix) <- rownames(adj_matrix) <- node_names
  }
  g <- igraph::graph_from_adjacency_matrix(t(adj_matrix), mode = "directed")
  igraph::plot.igraph(g, vertex.size = 20, vertex.color = "lightblue",
                      vertex.label.color = "black", vertex.label.cex = 0.8,
                      edge.arrow.size = 0.5, edge.color = "darkgray",
                      main = "Network Topology", ...)
  invisible(g)
}

#' Plot MCMC Trace for BBNI
#'
#' @description
#' Generates a trace plot of the log-posterior values over iterations of the MCMC
#' to visually assess convergence and stability of the executed Markov chain.
#' Burn-in line is graphed to show when the data started to be utilized for
#' edge-probability calculations.
#'
#' @param results The list returned by \code{run_bbni()}, containing \code{networks} and \code{log_posterior}.
#' @param every An integer specifying the thinning interval (sampling frequency) for plotting. Default is 1, which plots all log-posterior values. Values greater than 1 plot every \code{every}-th iteration.
#'
#' @return A base R trace plot.
#'
#' @examples
#' # 1. Generate synthetic network and time-series data
#' set.seed(123)
#' true_network <- GenerateNetwork(num.node = 5)
#' dummy_data <- GenerateSample(true_network, SampleSize = 15)
#'
#' # 2. Run BBNI sampler
#' prior_para <- matrix(3, nrow = 6, ncol = 2)
#' prior_para[6, 1] <- 2
#' prior_para[6, 2] <- 100
#'
#' results <- run_bbni(dummy_data, prior_para = prior_para, num_update = 100)
#'
#' # 3. Visualize MCMC results
#' plot_trace(results)
#'
#' @importFrom graphics plot abline legend
#' @export
plot_trace <- function(results, every = 1) {
  logpost <- results$log_posterior
  if (every > 1) logpost <- logpost[seq(1, length(logpost), by = every)]
  plot(logpost, type = "l", col = "darkblue",
      xlab = if (every == 1) "Node-level update" else "Outer iteration", 
      ylab = "Log-Posterior",
      main = "MCMC Trace Plot",
      lwd = 1.5)
  # add vertical line for burn-in if present
  if (!is.null(results$burn_in)) {
    abline(v = results$burn_in * length(logpost), col = "red", lty = 2)
    legend("bottomright", legend = paste("Burn-in =", results$burn_in), col = "red", lty = 2, bty = "n")
  }
}

#' Pre-compute Weighted Proposal Distribution
#'
#' Pre-computes a weighted proposal distribution for the MCMC algorithm by
#' evaluating all possible 1-input and 2-input Boolean logic functions using
#' Bayesian Information Criterion (BIC). The inverse BIC values of the potential
#' candidates are used as selection weights for the MCMC sampler.
#'
#' @param GeneData A matrix of observational gene expression data.
#' @param SampleSize An integer representing the total number of time points in the dataset.
#'
#' @return A list containing two matrices (`Candidate[[1]]` for 2-input triplets and `Candidate[[2]]` for 1-input pairs). These matrices store the candidate parent nodes, the best-fitting Boolean logic function, the raw miscounts, and the calculated proposal selection weights.
#' @noRd
ProposalConstruction <- function(GeneData, SampleSize, timeseries = TRUE) {
  gene.data <- GeneData
  error.prop <- 0.4
  pseudo.count <- 0.01
  orig_n <- SampleSize
  effective_n <- if (timeseries) orig_n - 1 else orig_n
  SampleSize <- effective_n
  threshold <- (SampleSize + pseudo.count * 8) * error.prop
  if (timeseries) {
    idx_parent <- 1:(orig_n - 1)
    idx_child <- 2:orig_n
  } else {
    idx_parent <- 1:orig_n
    idx_child <- 1:orig_n
  }
  candidate.prior <- list()
  kk <- 1
  for (i in seq_len(nrow(gene.data))) {
    for (j in seq_len(nrow(gene.data))) {
      if (j != i) {
        for (k in seq_len(nrow(gene.data))) {
          if (k != i && k != j) {
            test.result <- list()
            # vectorized frequency counting: maps (p1, p2, ch) binary states to integers 0-7
            triplet_states <- gene.data[i, idx_parent] * 4L + gene.data[j, idx_parent] * 2L + gene.data[k, idx_child] * 1L
            test.stat <- tabulate(triplet_states + 1L, nbins = 8)
            test.result[[1]] <- c(i, j, k, 1, BF1(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[2]] <- c(i, j, k, 2, BF2(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[3]] <- c(i, j, k, 3, BF3(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[4]] <- c(i, j, k, 4, BF4(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[5]] <- c(i, j, k, 5, BF5(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[6]] <- c(i, j, k, 6, BF6(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[7]] <- c(i, j, k, 7, BF7(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[8]] <- c(i, j, k, 8, BF8(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[9]] <- c(i, j, k, 9, BF9(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[10]] <- c(i, j, k, 10, BF10(test.stat, pseudo.count, SampleSize, threshold))
            test.result[[11]] <- c(i, j, k, 11, BF11(test.stat, pseudo.count, SampleSize, threshold)) # model g_k=g_i
            test.result[[12]] <- c(i, j, k, 12, BF12(test.stat, pseudo.count, SampleSize, threshold)) # model g_k=g_j
            test.result[[13]] <- c(i, j, k, 13, BF13(test.stat, pseudo.count, SampleSize, threshold)) # model g_k=complement(g_i)
            test.result[[14]] <- c(i, j, k, 14, BF14(test.stat, pseudo.count, SampleSize, threshold)) # model g_k=complement(g_j)
            miscount <- numeric()
            jj <- 1
            for (ii in seq_along(test.result)) {
              if (length(test.result[[ii]]) == 5) {
                miscount[jj] <- test.result[[ii]][5]
                jj <- jj + 1
              }
            }

            if (length(miscount) > 0) {
              min.miscount <- min(miscount)
              for (ii in seq_along(test.result)) {
                if (length(test.result[[ii]]) == 5 && test.result[[ii]][5] == min.miscount) {
                  candidate.prior[[kk]] <- test.result[[ii]]
                  kk <- kk + 1
                }
              }
            }
          }
        }
      }
    }
  }
  candidate <- matrix(nrow = length(candidate.prior), ncol = length(candidate.prior[[1]]))
  for (i in seq_along(candidate.prior)) {
    candidate[i, ] <- candidate.prior[[i]]
  }
  order.candidate <- candidate[order(candidate[, 3]), ] # order by output variables
  num.node <- max(order.candidate[, 3])
  triplet <- list()
  j1 <- 1
  pairwise <- list()
  j2 <- 1
  for (j in seq_len(nrow(order.candidate))) {
    if (order.candidate[j, 4] <= 10) {
      triplet[[j1]] <- order.candidate[j, ]
      j1 <- j1 + 1
    }
    if (order.candidate[j, 4] > 10) {
      if (order.candidate[j, 4] == 11) {
        pairwise[[j2]] <- c(order.candidate[j, 1], order.candidate[j, 3], 11, order.candidate[j, 5])
        j2 <- j2 + 1
      }
      if (order.candidate[j, 4] == 12) {
        pairwise[[j2]] <- c(order.candidate[j, 2], order.candidate[j, 3], 11, order.candidate[j, 5]) # model 11, g_k=g_i
        j2 <- j2 + 1
      }
      if (order.candidate[j, 4] == 13) {
        pairwise[[j2]] <- c(order.candidate[j, 1], order.candidate[j, 3], 12, order.candidate[j, 5])
        j2 <- j2 + 1
      }
      if (order.candidate[j, 4] == 14) {
        pairwise[[j2]] <- c(order.candidate[j, 2], order.candidate[j, 3], 12, order.candidate[j, 5]) # model 12: complement relation
        j2 <- j2 + 1
      }
    }
  }
  candidate.triplet <- matrix(nrow = length(triplet), ncol = 5)
  weighted.triplet <- matrix(nrow = length(triplet), ncol = 6)
  candidate.pairwise <- matrix(nrow = length(pairwise), ncol = 4)
  for (i in seq_along(triplet)) {
    candidate.triplet[i, ] <- triplet[[i]]
    for (j in 1:5) {
      weighted.triplet[i, j] <- candidate.triplet[i, j]
    }
    weighted.triplet[i, 6] <- candidate.triplet[i, 5]
  }
  for (i in seq_along(pairwise)) {
    candidate.pairwise[i, ] <- pairwise[[i]]
  }
  unique.pairwise <- unique(candidate.pairwise)
  weighted.pairwise <- matrix(nrow = length(unique.pairwise), ncol = 5)
  weighted.pairwise <- cbind(unique.pairwise, unique.pairwise[, 4])
  constant <- 0.001
  for (i in 1:num.node) {
    if (i %in% candidate.triplet[, 3]) {
      for (j in seq_len(nrow(candidate.triplet))) {
        if (candidate.triplet[j, 3] == i) {
          trip.matrix <- candidate.triplet[candidate.triplet[, 3] == i, 1:5]
          trip.matrix <- data.matrix(trip.matrix)
          if (ncol(trip.matrix) == 1) {
            trip.matrix <- t(trip.matrix)
          }
          score <- 1 / trip.matrix[, 5]
          prop <- (score + constant) / (sum(score + constant))
          trip.matrix[, 5] <- prop
          weighted.triplet[weighted.triplet[, 3] == i, ] <- cbind(trip.matrix, weighted.triplet[weighted.triplet[, 3] == i, 6])
        }
      }
    }
  }

  for (i in 1:num.node) {
    if (i %in% unique.pairwise[, 2]) {
      for (j in seq_len(nrow(unique.pairwise))) {
        if (unique.pairwise[j, 2] == i) {
          pair.matrix <- unique.pairwise[unique.pairwise[, 2] == i, 1:4]
          pair.matrix <- data.matrix(pair.matrix)
          if (ncol(pair.matrix) == 1) {
            pair.matrix <- t(pair.matrix)
          }
          score <- 1 / (pair.matrix[, 4] + constant)
          prop <- score / (sum(score))
          pair.matrix[, 4] <- prop
          weighted.pairwise[weighted.pairwise[, 2] == i, ] <- cbind(pair.matrix, weighted.pairwise[weighted.pairwise[, 2] == i, 5])
        }
      }
    }
  }
  CandidateTriplet <- weighted.triplet
  CandidatePairwise <- weighted.pairwise
  Candidate <- list()
  Candidate[[1]] <- CandidateTriplet
  Candidate[[2]] <- CandidatePairwise
  return(Candidate)
}

#' Build Initial Network Topology and Boolean Transition Matrix
#'
#' Builds a single valid starting network topology and Boolean transition function
#' matrix for the MCMC algorithm. It assigns parents using a 50/50 probability
#' of one versus two parents per node, then selects the most probable parent-child
#' relationships using pre-computed proposal weights. The function strictly ensures
#' the directed acyclic graph (DAG) constraint is maintained.
#'
#' @param Candidate A list containing transition function selection weights for 2-input and 1-input node combinations.
#' @param num.node An integer representing the total number of network nodes.
#'
#' @return A square transition function matrix representing the initial state of the network \eqn{T^{(0)}}{T^(0)} and its corresponding Boolean rules \eqn{F^{(0)}}{F^(0)}, providing an optimal acyclic starting network for the MCMC.
#' @importFrom stats runif
#' @noRd
ConstructInitial <- function(Candidate, num.node) {
  prior.triplet <- Candidate[[1]]
  prior.pairwise <- Candidate[[2]]
  trans_func_matrix <- matrix(0, nrow = num.node, ncol = num.node)
  prop_trans_func_matrix <- matrix(0, nrow = num.node, ncol = num.node)
  prop_incid_matrix <- matrix(0, nrow = num.node, ncol = num.node)
  incid_matrix <- matrix(0, nrow = num.node, ncol = num.node)
  ratio <- 0.5
  for (i in 1:num.node) {
    prop_trans_func_matrix <- trans_func_matrix
    prop_incid_matrix <- incid_matrix
    pairwise.prior.set <- matrix()
    pairwise.prior.set <- prior.pairwise[prior.pairwise[, 2] == i, ] # priors for node update_order[k]
    pairwise.prior.set <- data.matrix(pairwise.prior.set)
    if (ncol(pairwise.prior.set) == 1) {
      pairwise.prior.set <- t(pairwise.prior.set)
    }

    triplet.prior.set <- matrix()
    triplet.prior.set <- prior.triplet[prior.triplet[, 3] == i, ]
    triplet.prior.set <- data.matrix(triplet.prior.set)
    if (ncol(triplet.prior.set) == 1) {
      triplet.prior.set <- t(triplet.prior.set)
    }
    prop_prob <- runif(1)
    aa <- 0
    if (prop_prob >= ratio) { # consider two parents
      if (nrow(triplet.prior.set) >= 1) {
        candidate.pare.set <- list()
        for (jj in seq_len(nrow(triplet.prior.set))) {
          if (length(triplet.prior.set[jj, 1:2]) == 2) {
            aa <- aa + 1
            candidate.pare.set[[aa]] <- triplet.prior.set[jj, ]
          }
        }
        if (length(candidate.pare.set) > 0) {
          score <- numeric()
          for (jj in seq_along(candidate.pare.set)) {
            score[jj] <- candidate.pare.set[[jj]][5]
          }
          bbb <- matrix(nrow = length(candidate.pare.set), ncol = 6)
          for (jjj in seq_along(candidate.pare.set)) {
            bbb[jjj, ] <- candidate.pare.set[[jjj]]
          }

          max.score.candidate <- bbb[bbb[, 5] == max(score), ]
          if (is.numeric(max.score.candidate)) {
            add_parent <- max.score.candidate
          }
          if (is.matrix(max.score.candidate) && nrow(max.score.candidate) > 1) {
            add_parent <- max.score.candidate[sample.int(nrow(max.score.candidate), 1), ]
          }

          add_two_parent <- c(add_parent[1], add_parent[2])
        }
      }
      if (aa > 0) {
        func_order <- add_parent[4]
        prop_trans_func_matrix[i, add_two_parent[1]] <- func_order
        prop_trans_func_matrix[i, add_two_parent[2]] <- func_order
        for (ii in seq_len(nrow(prop_trans_func_matrix))) {
          for (jj in seq_len(ncol(prop_trans_func_matrix))) {
            if (prop_trans_func_matrix[ii, jj] > 0) {
              prop_incid_matrix[ii, jj] <- 1
            }
          }
        }
        prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
        if (check_ances_matrix(prop_ances_matrix) == 0) {
          trans_func_matrix[i, ] <- prop_trans_func_matrix[i, ]
          incid_matrix[i, ] <- prop_incid_matrix[i, ]
        }
      }
    }
    if (prop_prob < ratio && nrow(pairwise.prior.set) > 0) { # consider one parent
      candidate.pare.set <- pairwise.prior.set
      if (is.numeric(candidate.pare.set)) {
        add_parent <- candidate.pare.set
      }
      if (is.matrix(candidate.pare.set)) {
        add_parent <- candidate.pare.set[candidate.pare.set[, 4] == max(candidate.pare.set[, 4]), ]
      }
      if (is.matrix(add_parent) && nrow(add_parent) > 1) {
        add_parent <- add_parent[sample.int(nrow(add_parent), 1), ]
      }

      add_one_parent <- add_parent[1]

      func_order <- add_parent[3]
      prop_trans_func_matrix[i, add_one_parent] <- func_order
      for (ii in seq_len(nrow(prop_trans_func_matrix))) {
        for (jj in seq_len(ncol(prop_trans_func_matrix))) {
          if (prop_trans_func_matrix[ii, jj] > 0) {
            prop_incid_matrix[ii, jj] <- 1
          }
        }
      }
      prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
      if (check_ances_matrix(prop_ances_matrix) == 0) {
        trans_func_matrix[i, ] <- prop_trans_func_matrix[i, ]
        incid_matrix[i, ] <- prop_incid_matrix[i, ]
      }
    }
  }
  return(trans_func_matrix)
}

#' Calculate Transitive Closure of Adjacency Matrix
#'
#' Calculates the transitive closure of the network's directed adjacency matrix
#' to locate all ancestor-descendant relationships. This is used by the MCMC
#' algorithm to ensure proposed network topologies represent directed acyclic
#' graphs (DAGs) and to prevent illegal cyclic moves when adding or swapping parents.
#' @param incid_matrix A square binary adjacency matrix representing the current network topology.
#'
#' @returns A binary matrix of the same dimensions as `incid_matrix`. An entry of 1 at (i, k) indicates that node i is an ancestor of node k through one or more directed edges.
#' @noRd
update_ancestor_matrix <- function(incid_matrix) {
  n <- nrow(incid_matrix)
  R <- incid_matrix
  steps <- ceiling(log2(n))
  for (s in seq_len(steps)) {
    R <- (R %*% R + R) > 0
  }
  storage.mode(R) <- "double"
  return(R)
}

#' Check Ancestor Matrix for Cyclic Loops
#'
#' Ensures that the proposed network topology fulfills the directed acyclic graph
#' (DAG) constraint by checking the diagonal of the ancestor matrix for illegal
#' cyclic loops. This enforces the requirement that adding or swapping parents
#' cannot lead to a cyclic graph during MCMC topology updates.
#'
#' @param ances_matrix A binary transitive closure matrix of the network topology, where an entry of 1 at (i, j) means node i is an ancestor of node j.
#'
#' @return An integer count of the number of nodes that are ancestors of themselves. A value greater than 0 indicates the proposed topology has cyclic loops and must be rejected.
#' @noRd
check_ances_matrix <- function(ances_matrix) {
  # vectorized
  sum(diag(ances_matrix) == 1)
}

test_that("update_ancestor_matrix computes correct transitive closure", {
  # Chain: node 1 (root) -> node 2 -> node 3
  # incid_matrix[i, j] == 1 means node j is a parent of node i
  incid_matrix <- matrix(0, nrow = 3, ncol = 3)
  incid_matrix[2, 1] <- 1 # node 1 is a parent of node 2
  incid_matrix[3, 2] <- 1 # node 2 is a parent of node 3

  ances <- update_ancestor_matrix(incid_matrix)

  expect_equal(ances[2, 1], 1)
  expect_equal(ances[3, 2], 1)
  expect_equal(ances[3, 1], 1) # transitive: node 1 is an ancestor of node 3
  expect_equal(sum(diag(ances)), 0) # should be acyclic and have no self-ancestry
})

test_that("check_ances_matrix detects a directed cycle", {
  # 2-node cycle: 1 -> 2 -> 1
  incid_matrix <- matrix(c(0, 1, 1, 0), nrow = 2, byrow = TRUE)
  ances <- update_ancestor_matrix(incid_matrix)

  expect_gt(check_ances_matrix(ances), 0)
})

test_that("check_ances_matrix reports zero loops for a valid DAG", {
  incid_matrix <- matrix(0, nrow = 3, ncol = 3)
  incid_matrix[2, 1] <- 1
  incid_matrix[3, 2] <- 1
  ances <- update_ancestor_matrix(incid_matrix)

  expect_equal(check_ances_matrix(ances), 0)
})

test_that("GenerateNetwork always produces an acyclic topology", {
  set.seed(1)
  net <- GenerateNetwork(num.node = 8)
  incid <- (net > 0) * 1
  ances <- update_ancestor_matrix(incid)

  expect_equal(check_ances_matrix(ances), 0)
})

#' Generate Initial Network Topology
#'
#' Randomly generates a valid directed acyclic graph (DAG) topology
#' \eqn{T}{T} and assigns a corresponding Boolean transition function \eqn{F}{F} to each node.
#' The algorithm samples parent set configurations, keeping the constraint
#' that the maximum in-degree for any node is 2, and further ensures
#' the resulting structure does not contain directed cyclic loops.
#'
#' @param num.node An integer representing the total number of genes/nodes in the network.
#'
#' @return A square transition function matrix combining the initial DAG topology with randomly assigned Boolean logic functions. Elements with a value of 0 indicate no directed edge, while positive integers indicate the presence of an edge and specify the defining Boolean function type (codes 1-12: 1-10 for two-input functions, 11-12 for one-input identity/negation functions).
#'
#' @examples
#' # Generate a true network topology and Boolean rules for 5 nodes
#' set.seed(123)
#' true_network <- GenerateNetwork(num.node = 5)
#'
#' # Graph the network with built-in exported function plot_network
#' plot_network(true_network)
#'
#' @importFrom stats rbinom runif
#' @export
GenerateNetwork <- function(num.node) {
  # seems to end up into a nearly infinite loop at higher numbers of num.node, need to fix
  all_nodes <- 1:num.node
  loop <- 1
  while (loop != 0) {
    tent_incid_matrix <- matrix(0, nrow = num.node, ncol = num.node)
    tent_trans_matrix <- matrix(0, nrow = num.node, ncol = num.node)
    for (i in 1:num.node) {
      u1 <- runif(1)
      uu <- 4
      if (u1 > 1 / 10 && u1 <= uu / 10) {
        position <- sample(all_nodes[-i], 1)
        tent_incid_matrix[i, position] <- 1
        u2 <- runif(1)
        if (u2 > 0.5) {
          tent_trans_matrix[i, position] <- 11
        }
        if (u2 < 0.5) {
          tent_trans_matrix[i, position] <- 12
        }
      }
      if (u1 > uu / 10) {
        position <- sample(all_nodes[-i], 2)
        tent_incid_matrix[i, position[1]] <- 1
        tent_incid_matrix[i, position[2]] <- 1
        func <- sample(10, 1)
        tent_trans_matrix[i, position[1]] <- func
        tent_trans_matrix[i, position[2]] <- func
      }
    }
    tent_ances_matrix <- update_ancestor_matrix(tent_incid_matrix)
    loop <- check_ances_matrix(tent_ances_matrix)
  }
  return(tent_trans_matrix)
}
#' Simulate Boolean Network Observation Dataset
#'
#' Simulates a synthetic binary observation dataset (\eqn{G}{G}) based on a given
#' directed acyclic graph topology with Boolean rules. It starts by running
#' independent Bernoulli trials on root nodes. The remaining non-root nodes
#' are calculated based on their assigned Boolean logic functions and parent states.
#' A pre-generated binary noise matrix is applied via a bitwise XOR operation
#' to occasionally flip the Boolean outputs, injecting natural biological
#' noise expected by the model.
#'
#' @param trans_matrix A square matrix combining the network topology \eqn{T}{T} and integer-coded Boolean logic functions \eqn{F}{F} assigned to each directed edge.
#' @param SampleSize An integer representing the total number of samples or time points to simulate. Defaults to 50 if not specified.
#' @param num.node An integer representing the total number of network nodes. Defaults to \code{nrow(trans_matrix)} if not specified.
#' @param para A numeric vector of baseline success probabilities (\eqn{\theta_i}{\theta_i}) used to generate the expression states of root nodes via independent Bernoulli trials. Defaults to a \code{rep(0.5, nrow(trans_matrix))} if not specified.
#' @param error A pre-generated binary noise matrix applied to occasionally flip Boolean outputs, injecting natural noise. Defaults to a zero matrix of size \code{nrow(trans_matrix) x SampleSize} if not specified (no noise).
#' @param timeseries Logical. If TRUE, simulates time-lagged data where child state at time \eqn{t}{t} is determined by parent stated at \eqn{t-1}{t-1}. If FALSE, simulates independent samples where child and parent states are resolved simultaneously. Defaults to TRUE.
#'
#' @return A simulated binary gene expression matrix \eqn{G}{G}, where rows represent individual genes/nodes and columns represent samples.
#'
#' @examples
#' # 1. Generate a 5-node network
#' set.seed(123)
#' num_nodes <- 5
#' sample_size <- 10
#' true_network <- GenerateNetwork(num.node = num_nodes)
#'
#' # 2. Set baseline probabilities and simulate zero-noise error matrix
#' root_probs <- rep(0.5, num_nodes)
#' error_matrix <- matrix(0, nrow = num_nodes, ncol = sample_size)
#'
#' # 3. Generate cross-sectional (independent mode) synthetic data
#' dummy_data <- GenerateSample(
#'   trans_matrix = true_network,
#'   SampleSize = sample_size,
#'   para = root_probs,
#'   error = error_matrix,
#'   timeseries = FALSE
#' )
#' print(dummy_data)
#'
#' @export
GenerateSample <- function(trans_matrix, SampleSize = 50, num.node = nrow(trans_matrix), para = rep(0.5, nrow(trans_matrix)), error = matrix(0, nrow = nrow(trans_matrix), ncol = SampleSize), timeseries = TRUE) {
  node_ances <- matrix(nrow = num.node, ncol = 2)
  GeneData <- matrix(0, nrow = num.node, ncol = SampleSize)
  incid_matrix <- trans_matrix
  for (i in seq_len(nrow(trans_matrix))) {
    for (j in seq_len(ncol(trans_matrix))) {
      if (trans_matrix[i, j] > 0) {
        incid_matrix[i, j] <- 1
      }
    }
  }
  ances_matrix <- update_ancestor_matrix(incid_matrix)
  for (i in 1:num.node) {
    node_ances[i, 1] <- i
    node_ances[i, 2] <- sum(ances_matrix[i, ])
  }
  # sort nodes by ancestor count to evaluate in topological order
  node_ances <- node_ances[order(node_ances[, 2]), ]
  # dynamic idxing
  if (timeseries) {
    idx_child <- 2:SampleSize
    idx_parent <- 1:(SampleSize - 1)
  } else {
    idx_child <- 1:SampleSize
    idx_parent <- 1:SampleSize
  }
  for (i in seq_len(nrow(node_ances))) {
    if (node_ances[i, 2] == 0) {
      GeneData[node_ances[i, 1], ] <- rbinom(SampleSize, 1, prob = para[node_ances[i, 1]])
    } else {
      parent <- numeric()
      ii <- 1
      for (j in seq_len(ncol(incid_matrix))) {
        if (incid_matrix[node_ances[i, 1], j] != 0) {
          parent[ii] <- j
          ii <- ii + 1
        }
      }
      func <- trans_matrix[node_ances[i, 1], parent[1]]
      # bitwise boolean operations now mapped dynamically w/ idx_child and idx_parent
      if (func == 1) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(bitwAnd(GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 2) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(1 - bitwAnd(GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 3) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(bitwOr(GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 4) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(1 - bitwOr(GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 5) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(bitwOr(1 - GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 6) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(bitwOr(GeneData[parent[1], idx_parent], 1 - GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 7) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(bitwAnd(1 - GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 8) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(bitwAnd(GeneData[parent[1], idx_parent], 1 - GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 9) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(bitwXor(GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 10) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(1 - bitwXor(GeneData[parent[1], idx_parent], GeneData[parent[2], idx_parent]), error[node_ances[i, 1], idx_parent])
      }
      if (func == 11) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(GeneData[parent[1], idx_parent], error[node_ances[i, 1], idx_parent])
      }
      if (func == 12) {
        GeneData[node_ances[i, 1], idx_child] <- bitwXor(1 - GeneData[parent[1], idx_parent], error[node_ances[i, 1], idx_parent])
      }
    }
  }
  return(GeneData)
}

#' Yeast Cell-Cycle Gene Expression Data
#'
#' A binary empirical dataset with the gene expression states of 14 yeast
#' cell-cycle genes across 385 complete pooled experimental conditions. The
#' data were filtered to the 14 genes used by Han et al. (2014), columns with
#' missing values were removed, and each gene was binarized by its row mean.
#'
#' @format A numeric matrix with 14 rows (representing individual genes/nodes)
#' and 385 columns (representing independent samples or sequential time points).
#'
#' @source Han, S., Wong, R. K. W., Lee, T. C. M., Shen, L., Li, S.-Y. R., & Fan, X. (2014).
#' A Full Bayesian Approach for Boolean Genetic Network Inference. *PLOS ONE*, 9(12), e115806.
#' \doi{10.1371/journal.pone.0115806}
#'
#' @usage data(yeast_data)
"yeast_data"

#' Calculate Collapsed Posterior Log-Probability
#'
#' Calculates the collapsed posterior log-probability of a proposed network
#' topology (\eqn{T}{T}) and its Boolean functions (\eqn{F}{F}) given the observed data (\eqn{G}{G}).
#' It parses the transition function matrix (TRFUM) into root and non-root nodes,
#' predicts each node's value based on its Boolean rule, counts mismatches, and
#' evaluates the Beta-Binomial collapsed likelihood. By integrating out Bernoulli
#' parameters \eqn{\theta = \{p_E, p_i\}}{\theta = {p_E, p_i}} under Beta priors, the function computes
#' \eqn{\log p(T, F \mid G)}{\log p(T, F | G)}, the model score for the MCMC algorithm.
#'
#' @param TRFUM A matrix combining the current network topology and the specific Boolean logic functions assigned to each node (integer codes 1-12).
#' @param GeneData A matrix of the observed binary gene expression time-series dataset (rows = nodes, columns = time points).
#' @param SampleSize An integer representing the total number of time points/samples in `GeneData`.
#' @param num.node An integer representing the total number of genes/nodes in the network.
#' @param prior_para A matrix of Beta distribution hyperparameters (\eqn{\alpha, \beta}{\alpha, \beta}) for root nodes and the global noise parameter \eqn{e}{e}.
#' @param penalty Structural-prior hyperparameter in \eqn{(0, 1]}{(0, 1]}. A value of \code{1} corresponds to a uniform prior over valid network topologies; values below \code{1} apply an edge-count penalty that favors sparser networks.
#'
#' @return A list of two vectors evaluating model fit. `results[[1]]` contains `[ErrorFactor, RootFactor, likelihood, post_para, log_post_model]`, where `log_post_model` is the collapsed posterior metric \eqn{\log p(T, F \mid G)}{\log p(T, F | G)}. `results[[2]]` contains `[para_sample, mismatch, Perror]`, providing point estimates of the root ON-probabilities, total mismatches, and estimated noise error rate \eqn{e}{e}.
#' @noRd
Error_LLH <- function(TRFUM, GeneData, SampleSize, num.node, prior_para, penalty, timeseries = TRUE) {
  if (timeseries) {
    idx_child <- 2:SampleSize
    idx_parent <- 1:(SampleSize - 1)
  } else {
    idx_child <- 1:SampleSize
    idx_parent <- 1:SampleSize
  }
  n_obs <- length(idx_child)
  # speedup. vectorized row processing w/ pre-allocated memory instead of nested loops
  # drops time complexity from O(N^2) -> O(N)
  n <- nrow(TRFUM)
  row_has_parent <- rowSums(TRFUM != 0) > 0
  root_node <- which(!row_has_parent)
  nonroot_rows <- which(row_has_parent)
  InOutPair <- vector("list", length(nonroot_rows))
  for (idx in seq_along(nonroot_rows)) {
    i <- nonroot_rows[idx]
    jj <- which(TRFUM[i, ] != 0)
    p <- TRFUM[i, jj]
    InOutPair[[idx]] <- c(i, jj, p[length(p)])
  }
  n_nonroot <- length(InOutPair)
  mismatch <- 0
  gene_row_sums <- rowSums(GeneData)
  if (n_nonroot > 0) {
    for (i in seq_along(InOutPair)) {
      # reduce need of constant list indexing / length calls
      pair <- InOutPair[[i]]
      rule <- pair[length(pair)]
      out_node <- pair[1]
      child <- GeneData[out_node, idx_child]
      # handle node extraction based on the rule threshold
      if (rule <= 10) {
        parent1 <- GeneData[pair[2], idx_parent]
        parent2 <- GeneData[pair[3], idx_parent]
      } else {
        parent1 <- GeneData[pair[2], idx_parent]
      }
      # switch case
      mismatch <- mismatch + switch(rule,
      # 1
      sum(bitwXor(child, bitwAnd(parent1, parent2))),
      # 2
      sum(bitwXor(child, 1 - bitwAnd(parent1, parent2))),
      # 3
      sum(bitwXor(child, bitwOr(parent1, parent2))),
      # 4
      sum(bitwXor(child, 1 - bitwOr(parent1, parent2))),
      # 5
      sum(bitwXor(child, bitwOr(1 - parent1, parent2))),
      # 6
      sum(bitwXor(child, bitwOr(parent1, 1 - parent2))),
      # 7
      sum(bitwXor(child, bitwAnd(1 - parent1, parent2))),
      # 8
      sum(bitwXor(child, bitwAnd(parent1, 1 - parent2))),
      # 9
      sum(bitwXor(child, bitwXor(parent1, parent2))),
      # 10
      sum(bitwXor(child, 1 - bitwXor(parent1, parent2))),
      # 11
      sum(bitwXor(child, parent1)),
      # 12
      sum(bitwXor(child, 1 - parent1))
      )
    }
  }
  pseudo_count <- 0.0001
  raw_mismatch <- mismatch
  mismatch <- raw_mismatch + pseudo_count
  # use effective observation count for noise calculations
  if (n_nonroot > 0) {
    Perror <- mismatch / (n_nonroot * n_obs + pseudo_count)
    ErrorFactor <- mismatch * log(Perror) + (n_nonroot * n_obs - mismatch + pseudo_count) * log(1 - Perror)
    ErrorPrior <- (prior_para[num.node + 1, 1] - 1) * log(Perror) + (prior_para[num.node + 1, 2] - 1) * log(1 - Perror)
  } else {
    Perror <- NA
    ErrorFactor <- NA
    ErrorPrior <- 0
  }
  RootFactor <- numeric()
  RootPrior <- numeric()
  succ_count <- numeric()
  succ_prob <- numeric()
  for (i in seq_along(root_node)) {
    succ_count[i] <- sum(GeneData[root_node[i], ]) + pseudo_count
    succ_prob[i] <- succ_count[i] / (SampleSize + pseudo_count)
    RootFactor[i] <- succ_count[i] * log(succ_prob[i]) + (SampleSize + pseudo_count - succ_count[i]) * log(1 - succ_prob[i])

    RootPrior[i] <- (prior_para[root_node[i], 1] - 1) * log(succ_prob[i]) + (prior_para[root_node[i], 2] - 1) * log(1 - succ_prob[i])
  }
  if (n_nonroot == 0) { # all nodes are root nodes
    likelihood <- sum(RootFactor)
    post_para <- sum(RootFactor) + sum(RootPrior)
    log_post_model <- 0
    for (i in seq_len(nrow(TRFUM))) {
      nume <- lbeta(prior_para[i, 1] + gene_row_sums[i], prior_para[i, 2] + SampleSize - gene_row_sums[i])
      deno <- lbeta(prior_para[i, 1], prior_para[i, 2])
      log_post_model <- log_post_model + nume - deno
    }
    log_post_model <- log_post_model + length(TRFUM[TRFUM>0]) * log(penalty)
    ErrorFactor <- NA
    Perror <- NA
    mismatch <- NA
  }
  if (n_nonroot > 0) { # exist non root nodes
    likelihood <- ErrorFactor + sum(RootFactor)
    post_para <- ErrorFactor + sum(RootFactor) + sum(RootPrior) + ErrorPrior
    log_post_model <- 0
    for (i in seq_along(root_node)) {
      index <- root_node[i]
      nume <- lbeta(prior_para[index, 1] + gene_row_sums[index], prior_para[index, 2] + SampleSize - gene_row_sums[index])
      deno <- lbeta(prior_para[index, 1], prior_para[index, 2])
      log_post_model <- log_post_model + nume - deno
    }
    noise_nume <- lbeta(raw_mismatch + prior_para[num.node + 1, 1], n_nonroot * n_obs - raw_mismatch + prior_para[num.node + 1, 2])
    noise_deno <- lbeta(prior_para[num.node + 1, 1], prior_para[num.node + 1, 2])
    log_post_model <- log_post_model + noise_nume - noise_deno
    log_post_model <- log_post_model + length(TRFUM[TRFUM>0]) * log(penalty)
  }
  result <- list()
  result[[1]] <- c(ErrorFactor, RootFactor, likelihood, post_para, log_post_model)
  para_sample <- rep(NA, num.node)
  para_sample[root_node] <- succ_prob
  result[[2]] <- c(para_sample, mismatch, Perror)
  return(result)
}

#' Evaluate Boolean Function 1 (AND)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF1 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=g_i and g_j
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[2], test.stat[4], test.stat[6], test.stat[7])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 2 (NAND)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF2 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=complement(g_i and g_j)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[1], test.stat[3], test.stat[5], test.stat[8])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 3 (OR)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF3 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=(g_i or g_j)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[2], test.stat[3], test.stat[5], test.stat[7])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 4 (NOR)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF4 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=complement(g_i or g_j)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[1], test.stat[4], test.stat[6], test.stat[8])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 5 (OR-NOT)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF5 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=complement(g_i) or g_j
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[1], test.stat[3], test.stat[6], test.stat[7])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 6 (NOT-OR)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF6 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=g_i or complement(g_j)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[1], test.stat[4], test.stat[5], test.stat[7])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 7 (AND-NOT)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF7 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=complement(g_i) and g_j
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[2], test.stat[3], test.stat[6], test.stat[8])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 8 (NOT-AND)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF8 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=g_i and complement(g_j)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[2], test.stat[4], test.stat[5], test.stat[8])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 9 (XOR)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF9 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=g_i xor g_j
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[2], test.stat[3], test.stat[5], test.stat[8])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Function 10 (NXOR)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF10 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=complement(g_i xor g_j)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[1], test.stat[4], test.stat[6], test.stat[7])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Pairwise Relation 11 (Identity First Parent)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF11 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=g_i
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[2], test.stat[4], test.stat[5], test.stat[7])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Pairwise Relation 12 (Identity Second Parent)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF12 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=g_j
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[2], test.stat[3], test.stat[6], test.stat[7])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Negation 13 (NOT First Parent)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF13 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=complement(g_i)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[1], test.stat[3], test.stat[6], test.stat[8])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

#' Evaluate Boolean Negation 14 (NOT Second Parent)
#'
#' Computes a Bayesian Information Criterion (BIC) score for a candidate
#' Boolean function by counting state mismatches between predicted and
#' observed outputs. It compresses the raw data into a compact 3-gene
#' frequency count (truth table) to efficiently evaluate matching patterns.
#'
#' @param test.stat An integer frequency count vector storing how many times each specific combination of parent and child states occurs together in the data.
#' @param pseudo.count A numeric pseudocount added to avoid zero cells.
#' @param SampleSize An integer representing the total sample size.
#' @param threshold A numeric rejection threshold. If the error rate exceeds this, the function is rejected.
#'
#' @return A numeric BIC value, or `NULL` if the mismatch count exceeds the threshold.
#' @noRd
BF14 <- function(test.stat, pseudo.count, SampleSize, threshold) { # model g_k=complement(g_j)
  test.stat <- test.stat + pseudo.count
  false.count <- sum(test.stat[1], test.stat[4], test.stat[5], test.stat[8])
  error.estimate <- false.count / (SampleSize + pseudo.count * 8)
  BIC.value <- -2 * (false.count * log(error.estimate) + (SampleSize + pseudo.count * 8 - false.count) * log(1 - error.estimate)) + 2 * log(SampleSize)
  if (false.count <= threshold) {
    return(BIC.value)
  }
}

test_that("Error_LLH handles an all-root-node network correctly", {
  # A network with no edges: every node is a root node, exercising the
  # branch of Error_LLH that should report no error/mismatch terms.
  num.node <- 4
  SampleSize <- 50
  GeneData <- matrix(rbinom(num.node * SampleSize, 1, 0.5),
    nrow = num.node, ncol = SampleSize
  )
  TRFUM <- matrix(0, nrow = num.node, ncol = num.node)
  prior_para <- matrix(3, nrow = num.node + 1, ncol = 2)
  prior_para[num.node + 1, 1] <- 2
  prior_para[num.node + 1, 2] <- 100

  result <- Error_LLH(
    TRFUM = TRFUM, GeneData = GeneData, SampleSize = SampleSize,
    num.node = num.node, prior_para = prior_para, penalty = 0.1
  )

  expect_true(is.na(result[[1]][1])) # ErrorFactor: NA, no non-root nodes
  expect_true(is.na(result[[2]][num.node + 1])) # mismatch: NA
  expect_true(is.na(result[[2]][num.node + 2])) # Perror: NA

  log_post_model <- result[[1]][length(result[[1]])]
  expect_true(is.finite(log_post_model))
})

test_that("Error_LLH precisely shifts mismatch calculation for timeseries vs independent mode", {
  num.node <- 2
  SampleSize <- 4
  # Deterministic GeneData matrix:
  GeneData <- matrix(c(
    1, 0, 1, 0,
    0, 1, 0, 1
  ), nrow = num.node, byrow = TRUE)
  # Node 1 is the parent of Node 2 (Rule 11 = Identity).
  TRFUM <- matrix(0, nrow = num.node, ncol = num.node)
  TRFUM[2, 1] <- 11
  # Standard Beta priors
  prior_para <- matrix(3, nrow = num.node + 1, ncol = 2)
  prior_para[num.node + 1, 1] <- 2
  prior_para[num.node + 1, 2] <- 100
  # Test Time-Series Mode
  result_ts <- Error_LLH(
    TRFUM = TRFUM, GeneData = GeneData, SampleSize = SampleSize,
    num.node = num.node, prior_para = prior_para, penalty = 0.1,
    timeseries = TRUE
  )
  # Expect 0 actual mismatches + 0.0001 pseudo-count
  mismatch_ts <- result_ts[[2]][num.node + 1]
  expect_equal(mismatch_ts, 0.0001)
  # Test Independent Mode
  result_static <- Error_LLH(
    TRFUM = TRFUM, GeneData = GeneData, SampleSize = SampleSize,
    num.node = num.node, prior_para = prior_para, penalty = 0.1,
    timeseries = FALSE
  )
  # Expect 4 actual mismatches + 0.0001 pseudo-count
  mismatch_static <- result_static[[2]][num.node + 1]
  expect_equal(mismatch_static, 4.0001)
})

test_that("penalty = 1 produces uniform prior (no edge-count structural penalty)", {
  num.node <- 3
  SampleSize <- 10
  set.seed(42)
  GeneData <- matrix(rbinom(num.node * SampleSize, 1, 0.5),
                     nrow = num.node, ncol = SampleSize
  )
  prior_para <- matrix(3, nrow = num.node + 1, ncol = 2)
  prior_para[num.node + 1, 1] <- 2
  prior_para[num.node + 1, 2] <- 100
  # topology w/ 1 edge: node 2 has parent node 1 (identity function)
  TRFUM_1edge <- matrix(0, nrow = num.node, ncol = num.node)
  TRFUM_1edge[2, 1] <- 11
  # topology w/ 2 edges: node 2 has parent 1, node 3 has parent 1
  TRFUM_2edge <- matrix(0, nrow = num.node, ncol = num.node)
  TRFUM_2edge[2, 1] <- 11
  TRFUM_2edge[3, 1] <- 11
  # helper to get log_post_model (last element of first returned vector)
  get_logpost <- function(TRFUM, penalty) {
    res <- Error_LLH(
      TRFUM = TRFUM, GeneData = GeneData, SampleSize = SampleSize,
      num.node = num.node, prior_para = prior_para, penalty = penalty
    )
    res[[1]][length(res[[1]])]
  }
  # under penalty = 1, the structural term is d(T) * log(1) = 0 for all topologies
  # under penalty = 0.5, the structural term is d(T) * log(0.5)
  # therefore, logpost(pen=0.5) - logpost(pen=1) should equal d(T) * log(0.5)
  expect_equal(get_logpost(TRFUM_1edge, 0.5) - get_logpost(TRFUM_1edge, 1), 1 * log(0.5))
  expect_equal(get_logpost(TRFUM_2edge, 0.5) - get_logpost(TRFUM_2edge, 1), 2 * log(0.5))
  # under penalty = 1, the difference in log-posterior between two topologies
  # is purely from likelihood. the structural term contributes 0 to the difference.
  # under penalty = 0.5, the difference gains an extra (d2 - d1) * log(0.5) term.
  diff_unif <- get_logpost(TRFUM_2edge, 1) - get_logpost(TRFUM_1edge, 1)
  diff_pen  <- get_logpost(TRFUM_2edge, 0.5) - get_logpost(TRFUM_1edge, 0.5)
  expect_equal(diff_pen - diff_unif, (2 - 1) * log(0.5))
  # under penalty = 1, two topologies with the SAME edge count but different
  # edges should differ only by likelihood (structural term is identical: 0).
  TRFUM_2edge_alt <- matrix(0, nrow = num.node, ncol = num.node)
  TRFUM_2edge_alt[3, 2] <- 11
  TRFUM_2edge_alt[3, 1] <- 11
  diff_same_edge_count <- get_logpost(TRFUM_2edge, 1) - get_logpost(TRFUM_2edge_alt, 1)
  # structural terms cancel, so difference is finite and purely likelihood-based
  expect_true(is.finite(diff_same_edge_count))
})

#' Execute Metropolis-within-Gibbs MCMC Sampler for Boolean Networks
#'
#' Executes a Metropolis-within-Gibbs Markov chain Monte Carlo (MCMC) algorithm to sample
#' from the joint posterior distribution of Directed Acyclic Graph (DAG) topologies (\eqn{T}{T}) and
#' Boolean logic transition functions (\eqn{F}{F}). The algorithm iterates through individual
#' network nodes and proposes parent set mutations (edge additions, removals, or swaps)
#' paired with transition function reassignments to one of 14 candidate Boolean rules (stored as codes 1-12 in the transition matrix).
#' Proposed states transitions are strictly verified to follow the DAG constraint and
#' evaluated with a Metropolis-Hastings acceptance threshold using log-posterior values.
#'
#' @param GeneData A binary empirical observation matrix (\eqn{G}{G}), where rows represent individual network nodes (genes) and columns represent independent samples or sequential time points.
#' @param num.node An integer representing the total number of network nodes. Defaults to \code{nrow(GeneData)} if not specified.)
#' @param SampleSize An integer representing the total number of time points or independent samples in the dataset. Defaults to \code{ncol(GeneData)} if not specified.
#' @param prior_para A matrix (with dimensions `(num.node + 1) x 2`) of Beta prior hyperparameters \eqn{\alpha}{\alpha} and \eqn{\beta}{\beta} for root node probabilities and the global noise parameter \eqn{e}{e}. Defaults to a flat prior if not specified.
#' @param num_update An integer representing the total number of MCMC iterations to perform. Defaults to 4000 if not specified.
#' @param penalty Structural-prior hyperparameter in \eqn{(0, 1]}{(0, 1]}. A value of \code{1} corresponds to a uniform prior over valid network topologies; values below \code{1} apply an edge-count penalty that favors sparser networks. Defaults to 0.1 if not specified.
#' @param prop.ratio A numeric value between 0 and 1 giving the final probability of using the empirical proposal distribution after the first 10% of outer iterations. During the first 10% of outer iterations, the empirical proposal is used with probability 0.9. Defaults to 0.1.
#' @param verbose Logical. If TRUE, prints verbose MCMC iteration progress to the console. Default is FALSE.
#' @param timeseries Logical. If TRUE, the algorithm assumes a time-series dataset. If FALSE, the algorithm assumes independent samples. Default is TRUE.
#' @param burn_in A numeric value between 0 and 1 representing the proportion of initial MCMC samples to discard as burn-in. Defaults to 0.7 if not specified.
#'
#' @details Posterior edge probabilities (`post_edge_prob`) are computed from one
#' thinned sample per outer iteration (a full Gibbs sweep) after discarding
#' `burn_in`, as designed in the original method paper.
#'
#' @return A list containing the full trajectory of the MCMC chain. Specifically, `networks` (a list of sampled transition function matrices) and `log_posterior` (a numeric vector of log-posterior scores for each iteration). These represent samples drawn from the marginal posterior distribution \eqn{P(T,F|G)}{P(T,F|G)} used for Bayesian model averaging. Additionally, the `post_edge_prob` (matrix of marginal posterior edge probabilities) and `burn_in` ratio are returned in the list.
#'
#' @examples
#' # 1. Define network parameters
#' set.seed(235)
#' num_nodes <- 8
#' sample_size <- 50
#'
#' # 2. Generate true network and simulate data
#' true_network <- GenerateNetwork(num.node = num_nodes)
#'
#' # Set up Beta priors for root-node probabilities and the noise rate
#' prior_para <- matrix(3, nrow = num_nodes + 1, ncol = 2)
#' prior_para[num_nodes + 1, 1] <- 2
#' prior_para[num_nodes + 1, 2] <- 100
#'
#' # Simulate parameters
#' para <- numeric(num_nodes + 1)
#' for (i in 1:(num_nodes + 1)) {
#'   para[i] <- stats::rbeta(1, prior_para[i, 1], prior_para[i, 2])
#' }
#' para[num_nodes + 1] <- 0.1 # Fixed noise rate for simulation
#'
#' error_matrix <- matrix(stats::rbinom(num_nodes * sample_size, 1, para[num_nodes + 1]),
#'   nrow = num_nodes, ncol = sample_size
#' )
#'
#' dummy_data <- GenerateSample(
#'   trans_matrix = true_network,
#'   SampleSize = sample_size,
#'   para = para,
#'   error = error_matrix
#' )
#'
#' # 3. Run the MCMC sampler (silently)
#' mcmc_results <- run_bbni(
#'   GeneData = dummy_data,
#'   prior_para = prior_para,
#'   num_update = 80, # Scaled down for example speed
#'   prop.ratio = 0.1
#' )
#'
#' # 4. Visualize results
#' plot_bbni(mcmc_results, true_network = true_network, threshold = 0.5)
#' plot_trace(mcmc_results)
#'
#' @importFrom stats runif
#' @importFrom utils setTxtProgressBar txtProgressBar
#' @export
run_bbni <- function(GeneData, num.node = nrow(GeneData), SampleSize = ncol(GeneData), prior_para = NULL,
                     num_update = 4000, penalty = 0.1, prop.ratio = 0.1, verbose = FALSE, timeseries = TRUE, burn_in = 0.7) {
  # Generate flat priors if not provided by user
  if (is.null(prior_para)) {
    prior_para <- matrix(1, nrow = num.node + 1, ncol = 2)
  }
  # validate penalty: must be in (0, 1]
  if (!is.numeric(penalty) || length(penalty) != 1 || penalty <= 0 || penalty > 1) {
    stop("'penalty' must be a single numeric value in (0, 1]. Use penalty = 1 for a uniform topology prior.", call. = FALSE)
  }
  # validate prop.ratio
  if (!is.numeric(prop.ratio) || length(prop.ratio) != 1 || prop.ratio < 0 || prop.ratio > 1) {
    stop("'prop.ratio' must be a single numeric value in [0, 1].", call. = FALSE)
  }
  Candidate <- ProposalConstruction(GeneData, SampleSize, timeseries) # create the proposal for generated data
  prior.triplet <- Candidate[[1]]
  prior.pairwise <- Candidate[[2]]

  trans_func_matrix <- ConstructInitial(Candidate, num.node) # use the randomly selected initial
  incid_matrix <- trans_func_matrix
  for (i in seq_len(nrow(incid_matrix))) {
    for (j in seq_len(ncol(incid_matrix))) {
      if (incid_matrix[i, j] > 0) {
        incid_matrix[i, j] <- 1
      }
    }
  }
  ances_matrix <- update_ancestor_matrix(incid_matrix)


  Incidence_Matrix <- list()
  Ancestor_Matrix <- list() # Matrix: Ancestor Matrix recording ancestor-offspring relations for the whole chain
  Trans_Func_Matrix <- list() # Matrix: Transition Function Matrix for the whole chain
  Sample_Matrix <- matrix(nrow = num.node * num_update + 1, ncol = num.node + 2)
  Sample_Matrix[1, ] <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = trans_func_matrix, timeseries = timeseries)[[2]]
  Incidence_Matrix[[1]] <- incid_matrix
  Ancestor_Matrix[[1]] <- ances_matrix
  Trans_Func_Matrix[[1]] <- trans_func_matrix
  num <- numeric()
  logpost <- numeric()
  all_logpost <- numeric()
  aa <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = Trans_Func_Matrix[[1]], timeseries = timeseries)
  all_logpost[1] <- aa[[1]][length(aa[[1]])]
  logpost[1] <- all_logpost[1]
  n <- 1
  num[1] <- 1
  iter <- 1
  jump_point <- numeric()
  jump_point[1] <- 1 # parameters for each chain
  prop.ratio.final <- prop.ratio
  if (verbose) {
    # initialize progress bar
    cat("Running BBNI MCMC Sampling...\n")
    pb <- txtProgressBar(min = 0, max = num_update, style = 3)
  }
  for (ii in 1:num_update) { # run ii full rounds, with each round of num.node times
    if (ii <= round(0.1 * num_update)) { # use adaptive ratio of using proposal information
      prop.ratio <- 0.1
    } else {
      prop.ratio <- 1 - prop.ratio.final
    }
    
    update_order <- sample.int(num.node, num.node, replace = FALSE)
    for (k in seq_along(update_order)) { #   consider the updating node g_k
      if (verbose) {
        # update progress bar
        setTxtProgressBar(pb, ii)
      }
      old <- n
      current_incid_matrix <- Incidence_Matrix[[iter]]
      current_ances_matrix <- Ancestor_Matrix[[iter]]
      current_trans_func_matrix <- Trans_Func_Matrix[[iter]]
      current_post <- all_logpost[iter]

      parent_of_update <- numeric()
      j <- 1 # find the parent for node update_order[k]
      for (i in seq_len(ncol(current_incid_matrix))) {
        if (current_incid_matrix[update_order[k], i] != 0) {
          parent_of_update[j] <- i
          j <- j + 1
        }
      }
      swap_candi <- numeric()
      legal_parent <- numeric()
      j1 <- 1 # legal_parent is used for adding parent action
      for (i in 1:num.node) {
        if (i != update_order[k] && current_ances_matrix[i, update_order[k]] != 1 && !(i %in% parent_of_update)) {
          legal_parent[j1] <- i
          j1 <- j1 + 1
        }
      }
      swap_candi <- legal_parent
      pairwise.prior.set <- matrix() # clear variables
      pairwise.prior.pare <- matrix()
      pairwise.prior.set <- prior.pairwise[prior.pairwise[, 2] == update_order[k], ] # pairwise proposal for node update_order[k]
      pairwise.prior.set <- data.matrix(pairwise.prior.set)
      if (ncol(pairwise.prior.set) == 1) {
        pairwise.prior.set <- t(pairwise.prior.set)
        pairwise.prior.pare <- t(data.matrix(pairwise.prior.set[, 1]))
      }
      if (ncol(pairwise.prior.set) > 1 && nrow(pairwise.prior.set) > 0) {
        pairwise.prior.pare <- pairwise.prior.set[, 1]
      }

      triplet.prior.set <- matrix()
      triplet.prior.pare <- matrix()
      triplet.prior.set <- prior.triplet[prior.triplet[, 3] == update_order[k], ] # triplet proposal for node update_order[k]
      triplet.prior.set <- data.matrix(triplet.prior.set)
      if (ncol(triplet.prior.set) > 1 && nrow(triplet.prior.set) > 0) {
        triplet.prior.pare <- triplet.prior.set[, 1:2]
      }
      if (ncol(triplet.prior.set) == 1) {
        triplet.prior.set <- t(triplet.prior.set)
        triplet.prior.pare <- t(data.matrix(triplet.prior.set[, 1:2]))
      }
      pairwise.ratio <- 1 / 2
      # case one: no parent
      if (length(parent_of_update) == 0) { #  there are only add-parent moves, every time only one move can be proposed
        num_legal_parent <- length(legal_parent)
        if (num_legal_parent > 1) {
          uu <- runif(1)
        }
        if (num_legal_parent == 0) { # if no candidate nodes are available for adding parent, then ignore this node
          uu <- 0
        }
        if (num_legal_parent == 1 || (num_legal_parent > 1 && uu >= pairwise.ratio)) {
          # proposal move 1: add one parent
          prop.legal.overlap <- intersect(pairwise.prior.pare, legal_parent)

          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob >= prop.ratio && length(prop.legal.overlap) >= 1 && nrow(pairwise.prior.set) > 0) { # proposal information is used with probability 1-prop.ratio
            aa <- aa + 1
            candidate.pare.set <- pairwise.prior.set[pairwise.prior.set[, 1] %in% prop.legal.overlap, ]
            if (is.numeric(candidate.pare.set)) {
              add_parent <- candidate.pare.set
            }
            if (is.matrix(candidate.pare.set)) {
              add_parent <- candidate.pare.set[candidate.pare.set[, 4] == max(candidate.pare.set[, 4]), ]
            } # use the most likely one
            if (is.matrix(add_parent) && nrow(add_parent) > 1) {
              add_parent <- add_parent[sample.int(nrow(add_parent), 1), ]
            }

            add_one_parent <- add_parent[1]
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_node <- sample.int(length(legal_parent), 1, replace = FALSE)
            add_one_parent <- legal_parent[sample_node]
          }
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], add_one_parent] <- 1
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], add_one_parent] <- add_parent[3]
          } # function prior
          if (aa == 0) {
            func_order <- 10 + sample.int(2, 1)
            prop_trans_func_matrix[update_order[k], add_one_parent] <- func_order
          }
          add_one_prob <- 1 / length(legal_parent)
          prop_trans_func_prob <- 1 / 2 #  there are in all two types of boolean functions to choose for pairwise genes
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()
          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(1 / 2)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (num_legal_parent > 1) {
            p2c_prob[1] <- 1 / 3
          } # there are 3 possible moves, add one; remove one; swap one
          if (num_legal_parent == 1) {
            p2c_prob[1] <- 1
          }

          if (num_legal_parent > 1) {
            c2p_prob[1] <- 1 / 2 * add_one_prob
          } else {
            c2p_prob[1] <- add_one_prob
          }

          p2c_prob[2] <- 1
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        # proposal move 2 add two parents
        if (num_legal_parent > 1 && uu < pairwise.ratio) { # add two parents one time
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob >= prop.ratio && nrow(triplet.prior.set) >= 1) {
            candidate.pare.set <- list()
            for (jj in seq_len(nrow(triplet.prior.set))) {
              if (length(intersect(triplet.prior.set[jj, 1:2], legal_parent)) == 2) {
                aa <- aa + 1
                candidate.pare.set[[aa]] <- triplet.prior.set[jj, ]
              }
            }
            if (length(candidate.pare.set) > 0) {
              score <- numeric()
              for (jj in seq_along(candidate.pare.set)) {
                score[jj] <- candidate.pare.set[[jj]][5]
              }
              bbb <- matrix(nrow = length(candidate.pare.set), ncol = 6)
              for (jjj in seq_along(candidate.pare.set)) {
                bbb[jjj, ] <- candidate.pare.set[[jjj]]
              }

              max.score.candidate <- bbb[bbb[, 5] == max(score), ]
              if (is.numeric(max.score.candidate)) {
                add_parent <- max.score.candidate
              }
              if (is.matrix(max.score.candidate) && nrow(max.score.candidate) > 1) {
                add_parent <- max.score.candidate[sample.int(nrow(max.score.candidate), 1), ]
              }

              add_two_parent <- c(add_parent[1], add_parent[2])
            }
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_two_parent <- sample.int(length(legal_parent), 2, replace = FALSE)
            add_two_parent <- c(legal_parent[sample_two_parent[1]], legal_parent[sample_two_parent[2]])
          }
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], add_two_parent[1]] <- 1
          prop_incid_matrix[update_order[k], add_two_parent[2]] <- 1
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], add_two_parent[1]] <- add_parent[4]
            prop_trans_func_matrix[update_order[k], add_two_parent[2]] <- add_parent[4]
          }
          if (aa == 0) {
            func_order <- sample.int(10, 1)
            prop_trans_func_matrix[update_order[k], add_two_parent[1]] <- func_order
            prop_trans_func_matrix[update_order[k], add_two_parent[2]] <- func_order
          }
          sample_two_prob <- 1 / choose(num_legal_parent, 2)
          prop_trans_func_prob <- 1 / 10
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()
          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(swap_candi) > 1) {
            p2c_prob[1] <- 1 / 4
          } #  this is Q(T_c|T_p) inverse move, because there are 4 moves after adding two parents, depending on the number of nodes to swap
          if (length(swap_candi) == 1) {
            p2c_prob[1] <- 1 / 3
          }
          if (length(swap_candi) == 0) {
            p2c_prob[1] <- 1 / 2
          }
          c2p_prob[1] <- 1 / 2 * sample_two_prob # this is Q(T_p|T_c) there are adding one parent, adding two parents, two moves
          p2c_prob[2] <- 1
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
      }
      # case 2: one parent
      if (length(parent_of_update) == 1) {
        uu <- runif(1)
        one.parent.ratio <- 1 / 6
        # proposal move 1: add one parent
        if (length(legal_parent) > 0 && uu < one.parent.ratio) {
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob >= prop.ratio && nrow(triplet.prior.set) >= 1 && ncol(triplet.prior.set) > 1) {
            candidate.pare.set <- list()
            for (jj in seq_len(nrow(triplet.prior.set))) {
              if (length(intersect(parent_of_update, triplet.prior.pare[jj, ])) == 1 && parent_of_update %in% triplet.prior.pare[jj, ] && setdiff(triplet.prior.pare[jj, ], parent_of_update) %in% legal_parent) {
                aa <- aa + 1
                candidate.pare.set[[aa]] <- triplet.prior.set[jj, ]
              }
            }
            if (length(candidate.pare.set) > 0) {
              score <- numeric()
              for (jj in seq_along(candidate.pare.set)) {
                score[jj] <- candidate.pare.set[[jj]][5]
              }
              bbb <- matrix(nrow = length(candidate.pare.set), ncol = 6)
              for (jjj in seq_along(candidate.pare.set)) {
                bbb[jjj, ] <- candidate.pare.set[[jjj]]
              }

              max.score.candidate <- bbb[bbb[, 5] == max(score), ]
              if (is.numeric(max.score.candidate)) {
                add_parent <- max.score.candidate
              }
              if (is.matrix(max.score.candidate) && nrow(max.score.candidate) > 1) {
                add_parent <- max.score.candidate[sample.int(nrow(max.score.candidate), 1), ]
              }

              add_one_parent <- setdiff(add_parent[1:2], parent_of_update)
            }
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_node <- sample.int(length(legal_parent), 1, replace = FALSE)
            add_one_parent <- legal_parent[sample_node]
          }
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], add_one_parent] <- 1
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], add_one_parent] <- add_parent[4]
            prop_trans_func_matrix[update_order[k], parent_of_update] <- add_parent[4]
          }
          if (aa == 0) {
            func_order <- sample.int(10, 1)
            prop_trans_func_matrix[update_order[k], add_one_parent] <- func_order
            prop_trans_func_matrix[update_order[k], parent_of_update] <- func_order
          }
          add_one_prob <- 1 / length(legal_parent)
          prop_trans_func_prob <- 1 / 10 # there are 10 functions to choose
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()
          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(swap_candi) > 1) {
            p2c_prob[1] <- 1 / 4
          } #  this is Q(T_c|T_p) inverse move, because there are 4 moves after adding one parents, depending on the number of nodes to swap
          if (length(swap_candi) == 1) {
            p2c_prob[1] <- 1 / 3
          }
          if (length(swap_candi) == 0) {
            p2c_prob[1] <- 1 / 2
          }

          if (length(swap_candi) > 0) {
            c2p_prob[1] <- 1 / 3 * add_one_prob
          } else {
            c2p_prob[1] <- 1 / 2 * add_one_prob
          }

          p2c_prob[2] <- 1
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        # proposal move 2, swap one parent
        if (length(swap_candi) > 0 && uu >= 1 / 6 && uu < 2 / 6) {
          swap.prior.overlap <- intersect(swap_candi, pairwise.prior.pare)
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob > prop.ratio && length(swap.prior.overlap) > 0 && nrow(pairwise.prior.set) > 0) {
            aa <- aa + 1
            swap.candidate <- pairwise.prior.set[pairwise.prior.set[, 1] %in% swap.prior.overlap, ]
            if (is.numeric(swap.candidate)) { # i.e. swap.candidate has one row
              swap_parent <- swap.candidate
            }
            if (is.matrix(swap.candidate)) {
              swap_parent <- swap.candidate[swap.candidate[, 4] == max(swap.candidate[, 4]), ]
            } # multiple rows may have same weight

            if (is.matrix(swap_parent) && nrow(swap_parent) > 1) {
              swap_parent <- swap_parent[sample.int(nrow(swap_parent), 1), ]
            } # every candidate has equal chance
            swap_one_node <- swap_parent[1]
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_node <- sample.int(length(swap_candi), 1, replace = FALSE)
            swap_one_node <- swap_candi[sample_node]
          }
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], parent_of_update] <- 0
          prop_incid_matrix[update_order[k], swap_one_node] <- 1
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          prop_trans_func_matrix[update_order[k], parent_of_update] <- 0
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], swap_one_node] <- swap_parent[3]
          }
          if (aa == 0) {
            func_order <- 10 + sample.int(2, 1)
            prop_trans_func_matrix[update_order[k], swap_one_node] <- func_order
          }
          swap_one_prob <- 1 / length(swap_candi)
          prop_trans_func_prob <- 1 / 2 #  new proposal transition function matrix
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()
          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(legal_parent) > 0) {
            p2c_prob[1] <- 1 / 3
          } #  this is Q(T_c|T_p) inverse move, because there are 4 moves after adding one parents, depending on the number of nodes to swap
          if (length(legal_parent) == 0) {
            p2c_prob[1] <- 1 / 2
          }

          if (length(legal_parent) > 0) {
            c2p_prob[1] <- 1 / 3 * swap_one_prob
          } # this is Q(T_p|T_c) probability of which move and which node selected
          if (length(legal_parent) == 0) {
            c2p_prob[1] <- 1 / 2 * swap_one_prob
          }
          p2c_prob[2] <- 1
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        if (uu >= 2 / 6 && uu < 3 / 6) {
          remove_one_node <- parent_of_update
          remove_one_prob <- 1
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], remove_one_node] <- 0 # remove parent move does not need function.
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          prop_trans_func_matrix[update_order[k], remove_one_node] <- 0
          prop_trans_func_prob <- 1
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()
          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(legal_parent) == 0) {
            p2c_prob[1] <- 1
          }
          if (length(legal_parent) == 1) {
            p2c_prob[1] <- 1
          } #  this is Q(T_c|T_p) inverse move
          if (length(legal_parent) > 1) {
            p2c_prob[1] <- 1 / 2
          }

          if (length(legal_parent) > 0) {
            c2p_prob[1] <- 1 / 3 * remove_one_prob
          } # this is Q(T_p|T_c) probability of which move and which node selected
          if (length(legal_parent) == 0) {
            c2p_prob[1] <- 1 / 2 * remove_one_prob
          }
          p2c_prob[2] <- 1 / 2
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        # additional move 1: reverse one arc in pairwise genes
        if (uu >= 3 / 6 && uu < 4 / 6) {
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], parent_of_update] <- 0
          prop_incid_matrix[parent_of_update, update_order[k]] <- 1
          parent_parent <- numeric()
          j <- 0
          for (i in seq_len(nrow(current_incid_matrix))) {
            if (current_incid_matrix[parent_of_update, i] > 0) {
              j <- j + 1
              parent_parent[j] <- i
            }
          }
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          if (check_ances_matrix(prop_ances_matrix) == 0 && length(parent_parent) == 0) { # make sure no loops && parent_of_update no parents
            prop_trans_func_matrix <- current_trans_func_matrix
            func_order <- current_trans_func_matrix[update_order[k], parent_of_update]
            prop_trans_func_matrix[update_order[k], parent_of_update] <- 0
            prop_trans_func_matrix[parent_of_update, update_order[k]] <- func_order
            xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
            prop_sample <- xxx[[2]]
            prop_post <- xxx[[1]][length(xxx[[1]])]

            nume <- sum((prop_post))
            deno <- sum((current_post))
            acce_prob <- exp(nume - deno)
            ratio <- runif(1)
            if (ratio <= acce_prob) {
              n <- n + 1
            }
          }
        }
        # additional move 2: reverse two arcs simultanously
        if (uu >= 4 / 6 && uu < 5 / 6) {
          parent_parent <- numeric()
          j1 <- 0
          children <- numeric()
          j <- 0 # find the children of parent_of_update
          for (i in seq_len(ncol(current_incid_matrix))) {
            if (i != update_order[k] && current_incid_matrix[i, parent_of_update] > 0) {
              j <- j + 1
              children[j] <- i
            }
            if (current_incid_matrix[parent_of_update, i] > 0) {
              j1 <- j1 + 1
              parent_parent[j1] <- i
            }
          }
          if (length(children) == 1 && length(parent_parent) == 0) { # parent_of_update has two exact 2 children and no parents
            parent_child <- numeric()
            j <- 0 # find parent of children to determine the trans_func_matrix
            for (i in 1:num.node) {
              if (current_incid_matrix[children, i] > 0) {
                j <- j + 1
                parent_child[j] <- i
              }
            }

            prop_incid_matrix <- current_incid_matrix
            prop_incid_matrix[update_order[k], parent_of_update] <- 0
            prop_incid_matrix[children, parent_of_update] <- 0
            prop_incid_matrix[parent_of_update, update_order[k]] <- 1
            prop_incid_matrix[parent_of_update, children] <- 1
            prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
            if (check_ances_matrix(prop_ances_matrix) == 0) { # ensure no directed cycles
              prop_trans_func_matrix <- current_trans_func_matrix
              prop_trans_func_matrix[update_order[k], parent_of_update] <- 0
              prop_trans_func_matrix[children, parent_of_update] <- 0
              func_order <- sample.int(10, 1)
              prop_trans_func_matrix[parent_of_update, update_order[k]] <- func_order
              prop_trans_func_matrix[parent_of_update, children] <- func_order
              if (length(parent_child) == 2) {
                prop_trans_func_matrix[children, setdiff(parent_child, parent_of_update)] <- 10 + sample.int(2, 1)
              }
              xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
              prop_sample <- xxx[[2]]
              prop_post <- xxx[[1]][length(xxx[[1]])]

              nume <- sum((prop_post))
              deno <- sum((current_post))
              acce_prob <- exp(nume - deno)
              ratio <- runif(1)
              if (ratio <= acce_prob) {
                n <- n + 1
              }
            }
          }
        }
        if (uu >= 5 / 6 && length(swap_candi) > 1) {
          legal.triplet.pare <- intersect(swap_candi, triplet.prior.pare)
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob > prop.ratio) {
            swap.candidate <- list()
            if (length(intersect(parent_of_update, legal.triplet.pare)) == 0 && nrow(triplet.prior.set) > 0) {
              for (jj in seq_len(nrow(triplet.prior.pare))) {
                if (length(intersect(parent_of_update, triplet.prior.pare[jj, ])) == 0 && length(intersect(triplet.prior.pare[jj, ], swap_candi)) == 2) {
                  aa <- aa + 1
                  swap.candidate[[aa]] <- triplet.prior.set[jj, ]
                }
              }
            }
            if (length(swap.candidate) > 0) {
              score <- numeric()
              for (jj in seq_along(swap.candidate)) {
                score[jj] <- swap.candidate[[jj]][5]
              }
              bbb <- matrix(nrow = length(swap.candidate), ncol = 6) # convert list to matrix
              for (jjj in seq_along(swap.candidate)) {
                bbb[jjj, ] <- swap.candidate[[jjj]]
              }
              max.score.candidate <- bbb[bbb[, 5] == max(score), ]
              if (is.numeric(max.score.candidate)) {
                swap_parent <- max.score.candidate
              }
              if (is.matrix(max.score.candidate) && nrow(max.score.candidate) > 1) {
                swap_parent <- max.score.candidate[sample.int(nrow(max.score.candidate), 1), ]
              }
              swap_two_parent <- swap_parent[1:2]
            }
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_two_node <- sample.int(length(swap_candi), 2, replace = FALSE)
            swap_two_parent <- c(swap_candi[sample_two_node[1]], swap_candi[sample_two_node[2]])
          }
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], parent_of_update] <- 0
          prop_incid_matrix[update_order[k], swap_two_parent[1]] <- 1
          prop_incid_matrix[update_order[k], swap_two_parent[2]] <- 1
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          prop_trans_func_matrix[update_order[k], parent_of_update] <- 0
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], swap_two_parent[1]] <- swap_parent[4]
            prop_trans_func_matrix[update_order[k], swap_two_parent[2]] <- swap_parent[4]
          }
          if (aa == 0) {
            func_order <- sample.int(10, 1)
            prop_trans_func_matrix[update_order[k], swap_two_parent[1]] <- func_order
            prop_trans_func_matrix[update_order[k], swap_two_parent[2]] <- func_order
          }
          swap_two_prob <- 1 / choose(length(swap_candi), 2)
          prop_trans_func_prob <- 1 / 10
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()

          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(swap_candi) > 1) {
            p2c_prob[1] <- 1 / 4
          } #  this is Q(T_c|T_p) inverse move, because there are 4 moves after adding one parents, depending on the number of nodes to swap
          if (length(swap_candi) == 1) {
            p2c_prob[1] <- 1 / 3
          }

          if (length(swap_candi) > 1) {
            c2p_prob[1] <- 1 / 4 * swap_two_prob
          } # this is Q(T_p|T_c) probability of which move and which node selected
          if (length(swap_candi) == 1) {
            c2p_prob[1] <- 1 / 3 * swap_two_prob
          }
          p2c_prob[2] <- 1 / 10
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
      }
      # case3: two parents
      if (length(parent_of_update) == 2) {
        uu <- runif(1)
        two.parent.ratio <- 1 / 7
        # proposal move 1: remove one parent
        if (uu < two.parent.ratio) {
          pare.prior.overlap <- intersect(parent_of_update, pairwise.prior.pare)
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob >= prop.ratio && length(pare.prior.overlap) == 1 && nrow(pairwise.prior.set) > 0) {
            aa <- aa + 1
            remove_one_node <- setdiff(parent_of_update, pairwise.prior.pare)
            remove_pare <- pairwise.prior.set[pairwise.prior.set[, 1] == setdiff(parent_of_update, remove_one_node), ]
            if (is.numeric(remove_pare)) {
              func_order <- remove_pare[3]
            }
            if (is.matrix(remove_pare)) {
              Remove_Pare <- remove_pare[remove_pare[, 4] == max(remove_pare[, 4]), ]
              if (is.matrix(Remove_Pare) && nrow(Remove_Pare) > 1) {
                Remove_Pare <- Remove_Pare[sample.int(nrow(Remove_Pare), 1), ]
              } # each candidate has equal chance
              func_order <- Remove_Pare[3]
            }
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_node <- sample.int(length(parent_of_update), 1, replace = FALSE)
            remove_one_node <- parent_of_update[sample_node]
          }
          remove_one_prob <- 1 / length(parent_of_update)
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], remove_one_node] <- 0
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          prop_trans_func_matrix[update_order[k], remove_one_node] <- 0
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], setdiff(parent_of_update, remove_one_node)] <- func_order
          } # by the definition
          if (aa == 0) {
            func_order <- 10 + sample.int(2, 1)
          }
          prop_trans_func_matrix[update_order[k], setdiff(parent_of_update, remove_one_node)] <- func_order
          prop_trans_func_prob <- 1 / 2
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()

          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(swap_candi) > 0) {
            p2c_prob[1] <- 1 / 3
          } #  this is Q(T_c|T_p) inverse move, because there are 4 moves after adding one parents, depending on the number of nodes to swap
          if (length(swap_candi) == 0) {
            p2c_prob[1] <- 1 / 2
          }

          if (length(swap_candi) == 0) {
            c2p_prob[1] <- 1 / 2
          }
          if (length(swap_candi) > 1) {
            c2p_prob[1] <- 1 / 4 * remove_one_prob
          } # this is Q(T_p|T_c) probability of which move and which node selected
          if (length(swap_candi) == 1) {
            c2p_prob[1] <- 1 / 3 * remove_one_prob
          }

          p2c_prob[2] <- 1 / 10
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        # proposal move 2; remove two parents
        if (uu >= two.parent.ratio && uu < 2 / 7) {
          remove_node <- parent_of_update
          remove_prob <- 1
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], remove_node[1]] <- 0
          prop_incid_matrix[update_order[k], remove_node[2]] <- 0
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          prop_trans_func_matrix[update_order[k], remove_node[1]] <- 0
          prop_trans_func_matrix[update_order[k], remove_node[2]] <- 0
          prop_trans_func_prob <- 1
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()
          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          p2c_prob[1] <- 1 / 2 #  this is Q(T_c|T_p) inverse move
          c2p_prob[1] <- 1 / 4 * remove_prob # this is Q(T_p|T_c) probability of which move and which node selected
          p2c_prob[2] <- 1 / 10
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        # proposal move 3: swap one parent
        if (uu >= 2 / 7 && uu < 3 / 7 && length(swap_candi) > 0) {
          legal.triplet.pare <- intersect(swap_candi, triplet.prior.pare)
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob >= prop.ratio) {
            swap.candidate <- list()
            if (length(intersect(parent_of_update, legal.triplet.pare)) == 1 && nrow(triplet.prior.set) > 0) {
              for (jj in seq_len(nrow(triplet.prior.pare))) {
                if (length(intersect(parent_of_update, triplet.prior.pare[jj, ])) == 1) {
                  aa <- aa + 1
                  swap.candidate[[aa]] <- triplet.prior.set[jj, ]
                }
              }
            }
            if (length(swap.candidate) > 0) {
              score <- numeric()
              for (jj in seq_along(swap.candidate)) {
                score[jj] <- swap.candidate[[jj]][5]
              }
              bbb <- matrix(nrow = length(swap.candidate), ncol = 6)
              for (jjj in seq_along(swap.candidate)) {
                bbb[jjj, ] <- swap.candidate[[jjj]]
              }

              max.score.candidate <- bbb[bbb[, 5] == max(score), ]
              if (is.numeric(max.score.candidate)) {
                swap_parent <- max.score.candidate
              }
              if (is.matrix(max.score.candidate) && nrow(max.score.candidate) > 1) {
                swap_parent <- max.score.candidate[sample.int(nrow(max.score.candidate), 1), ]
              }
              swap_one_node <- setdiff(swap_parent[1:2], parent_of_update)
              sample_one_parent <- setdiff(parent_of_update, swap_parent[1:2])
            }
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_node <- sample.int(length(swap_candi), 1, replace = FALSE)
            sample_parent <- sample.int(2, 1, replace = FALSE)
            swap_one_node <- swap_candi[sample_node]
            sample_one_parent <- parent_of_update[sample_parent]
          }
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], sample_one_parent] <- 0
          prop_incid_matrix[update_order[k], swap_one_node] <- 1
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          prop_trans_func_matrix[update_order[k], sample_one_parent] <- 0
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], setdiff(parent_of_update, sample_one_parent)] <- swap_parent[4]
            prop_trans_func_matrix[update_order[k], swap_one_node] <- swap_parent[4]
          }
          if (aa == 0) {
            func_order <- sample.int(10, 1)
            prop_trans_func_matrix[update_order[k], setdiff(parent_of_update, sample_one_parent)] <- func_order
            prop_trans_func_matrix[update_order[k], swap_one_node] <- func_order
          }
          swap_one_prob <- 1 / length(swap_candi) * 1 / 2
          prop_trans_func_prob <- 1 / 10
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()
          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(swap_candi) == 0) {
            p2c_prob[1] <- 1
          } # since we use log(p2c_prob)
          if (length(swap_candi) > 1) {
            p2c_prob[1] <- 1 / 4
          } #  this is Q(T_c|T_p) inverse move, because there are 4 moves after adding one parents, depending on the number of nodes to swap
          if (length(swap_candi) == 1) {
            p2c_prob[1] <- 1 / 3
          }

          if (length(swap_candi) == 0) {
            c2p_prob[1] <- 1
          }
          if (length(swap_candi) > 1) {
            c2p_prob[1] <- 1 / 4 * swap_one_prob
          } # this is Q(T_p|T_c) probability of which move and which node selected
          if (length(swap_candi) == 1) {
            c2p_prob[1] <- 1 / 3 * swap_one_prob
          }
          p2c_prob[2] <- 1 / 10
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        # proposal move 4, swap two parents
        if (uu >= 3 / 7 && uu < 4 / 7 && length(swap_candi) > 1) {
          legal.triplet.pare <- intersect(swap_candi, triplet.prior.pare)
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob > prop.ratio) {
            swap.candidate <- list()
            if (length(intersect(parent_of_update, legal.triplet.pare)) == 0 && nrow(triplet.prior.set) > 0) {
              for (jj in seq_len(nrow(triplet.prior.pare))) {
                if (length(intersect(parent_of_update, triplet.prior.pare[jj, ])) == 0 && length(intersect(triplet.prior.pare[jj, ], swap_candi)) == 2) {
                  aa <- aa + 1
                  swap.candidate[[aa]] <- triplet.prior.set[jj, ]
                }
              }
            }
            if (length(swap.candidate) > 0) {
              score <- numeric()
              for (jj in seq_along(swap.candidate)) {
                score[jj] <- swap.candidate[[jj]][5]
              }
              bbb <- matrix(nrow = length(swap.candidate), ncol = 6) # convert list to matrix
              for (jjj in seq_along(swap.candidate)) {
                bbb[jjj, ] <- swap.candidate[[jjj]]
              }
              max.score.candidate <- bbb[bbb[, 5] == max(score), ]
              if (is.numeric(max.score.candidate)) {
                swap_parent <- max.score.candidate
              }
              if (is.matrix(max.score.candidate) && nrow(max.score.candidate) > 1) {
                swap_parent <- max.score.candidate[sample.int(nrow(max.score.candidate), 1), ]
              }
              swap_two_parent <- swap_parent[1:2]
            }
          }
          if (prop.prob < prop.ratio || aa == 0) {
            sample_two_node <- sample.int(length(swap_candi), 2, replace = FALSE)
            swap_two_parent <- c(swap_candi[sample_two_node[1]], swap_candi[sample_two_node[2]])
          }
          prop_incid_matrix <- current_incid_matrix
          prop_incid_matrix[update_order[k], parent_of_update[1]] <- 0
          prop_incid_matrix[update_order[k], parent_of_update[2]] <- 0
          prop_incid_matrix[update_order[k], swap_two_parent[1]] <- 1
          prop_incid_matrix[update_order[k], swap_two_parent[2]] <- 1
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          prop_trans_func_matrix[update_order[k], parent_of_update[1]] <- 0
          prop_trans_func_matrix[update_order[k], parent_of_update[2]] <- 0
          if (aa > 0) {
            prop_trans_func_matrix[update_order[k], swap_two_parent[1]] <- swap_parent[4]
            prop_trans_func_matrix[update_order[k], swap_two_parent[2]] <- swap_parent[4]
          }
          if (aa == 0) {
            func_order <- sample.int(10, 1)
            prop_trans_func_matrix[update_order[k], swap_two_parent[1]] <- func_order
            prop_trans_func_matrix[update_order[k], swap_two_parent[2]] <- func_order
          }
          swap_two_prob <- 1 / choose(length(swap_candi), 2)
          prop_trans_func_prob <- 1 / 10
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          prop_sample_prob <- numeric()
          p2c_prob <- numeric() # calculate acceptance probability
          curr_sample_prob <- numeric()
          c2p_prob <- numeric()

          prop_sample_prob[1] <- prop_post
          prop_sample_prob[2] <- 0 #  this is the prior && likelihood
          curr_sample_prob[1] <- current_post[1]
          curr_sample_prob[2] <- 0
          prop_sample_prob[3] <- log(prop_trans_func_prob)
          curr_sample_prob[3] <- log(1) #  this is P(R|T)

          if (length(swap_candi) > 1) {
            p2c_prob[1] <- 1 / 4
          } #  this is Q(T_c|T_p) inverse move, because there are 4 moves after adding one parents, depending on the number of nodes to swap
          if (length(swap_candi) == 1) {
            p2c_prob[1] <- 1 / 3
          }

          if (length(swap_candi) > 1) {
            c2p_prob[1] <- 1 / 4 * swap_two_prob
          } # this is Q(T_p|T_c) probability of which move and which node selected
          if (length(swap_candi) == 1) {
            c2p_prob[1] <- 1 / 3 * swap_two_prob
          }
          p2c_prob[2] <- 1 / 10
          c2p_prob[2] <- prop_trans_func_prob
          p2c_prob[3] <- 1
          c2p_prob[3] <- 1

          nume <- sum((prop_sample_prob)) + sum(log(p2c_prob))
          deno <- sum((curr_sample_prob)) + sum(log(c2p_prob))
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (ratio <= acce_prob) {
            n <- n + 1
          }
        }
        #  additional move 1: reorder the input and output variables  for XOR relation
        if (uu >= 4 / 7 && uu < 5 / 7 && current_trans_func_matrix[update_order[k], parent_of_update[1]] == 9) {
          prop_incid_matrix <- current_incid_matrix
          sample.child <- parent_of_update[sample.int(2, 1)]
          for (i in 1:2) {
            prop_incid_matrix[update_order[k], parent_of_update[i]] <- 0
          }
          parent_parent <- numeric()
          mm <- 0
          for (i in 1:num.node) {
            if (current_incid_matrix[sample.child, i] > 0) {
              mm <- mm + 1
              parent_parent[mm] <- i
            }
          }
          if (length(parent_parent) == 0) { # selected child should have no parents before this operation
            prop_incid_matrix[sample.child, update_order[k]] <- 1
            prop_incid_matrix[sample.child, setdiff(parent_of_update, sample.child)] <- 1
            prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
            if (check_ances_matrix(prop_ances_matrix) == 0) { # make sure no directed cycles
              prop_trans_func_matrix <- current_trans_func_matrix
              func_order <- current_trans_func_matrix[update_order[k], parent_of_update[1]]
              for (i in 1:2) {
                prop_trans_func_matrix[update_order[k], parent_of_update[i]] <- 0
              }
              prop_trans_func_matrix[sample.child, update_order[k]] <- func_order
              prop_trans_func_matrix[sample.child, setdiff(parent_of_update, sample.child)] <- func_order
              xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
              prop_sample <- xxx[[2]]
              prop_post <- xxx[[1]][length(xxx[[1]])]

              nume <- sum((prop_post))
              deno <- sum((current_post))
              acce_prob <- exp(nume - deno)
              ratio <- runif(1)
              if (ratio <= acce_prob) {
                n <- n + 1
              }
            }
          }
        }
        # additional move 2: reverse one arc among gene triplet
        if (uu >= 5 / 7 && uu < 6 / 7) {
          reverse_pare <- parent_of_update[sample.int(2, 1)]
          remain_pare <- setdiff(parent_of_update, reverse_pare)
          reverse_node_pare <- numeric()
          j <- 1 # find the already existing parent for reverse_pare
          for (i in seq_len(ncol(current_incid_matrix))) {
            if (current_incid_matrix[reverse_pare, i] != 0) {
              reverse_node_pare[j] <- i
              j <- j + 1
            }
          }
          if (length(reverse_node_pare) == 0) { # currently no parent
            prop_incid_matrix <- current_incid_matrix
            prop_incid_matrix[update_order[k], reverse_pare] <- 0
            prop_incid_matrix[reverse_pare, update_order[k]] <- 1
            prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
            if (check_ances_matrix(prop_ances_matrix) == 0) { # make sure no loops
              prop_trans_func_matrix <- current_trans_func_matrix
              prop_trans_func_matrix[update_order[k], reverse_pare] <- 0
              prop_trans_func_matrix[update_order[k], remain_pare] <- 10 + sample.int(2, 1)
              prop_trans_func_matrix[reverse_pare, update_order[k]] <- 10 + sample.int(2, 1)
              xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
              prop_sample <- xxx[[2]]
              prop_post <- xxx[[1]][length(xxx[[1]])]

              nume <- sum((prop_post))
              deno <- sum((current_post))
              acce_prob <- exp(nume - deno)
              ratio <- runif(1)
              if (ratio <= acce_prob) {
                n <- n + 1
              }
            }
          }
          if (length(reverse_node_pare) == 1) { # currently one parent
            prop_incid_matrix <- current_incid_matrix
            prop_incid_matrix[update_order[k], reverse_pare] <- 0
            prop_incid_matrix[reverse_pare, update_order[k]] <- 1
            prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
            if (check_ances_matrix(prop_ances_matrix) == 0) { # make sure no loops
              prop_trans_func_matrix <- current_trans_func_matrix
              prop_trans_func_matrix[update_order[k], reverse_pare] <- 0
              prop_trans_func_matrix[update_order[k], remain_pare] <- 10 + sample.int(2, 1)
              func_order <- sample.int(10, 1)
              prop_trans_func_matrix[reverse_pare, update_order[k]] <- func_order
              prop_trans_func_matrix[reverse_pare, reverse_node_pare] <- func_order
              xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
              prop_sample <- xxx[[2]]
              prop_post <- xxx[[1]][length(xxx[[1]])]

              nume <- sum((prop_post))
              deno <- sum((current_post))
              acce_prob <- exp(nume - deno)
              ratio <- runif(1)
              if (ratio <= acce_prob) {
                n <- n + 1
              }
            }
          }
        }
        # additional move 3: only change boolean functions among gene triplets  note: the network may remain the same
        if (uu >= 6 / 7) {
          prop.prob <- runif(1)
          aa <- 0
          if (prop.prob >= prop.ratio && nrow(triplet.prior.set) >= 1 && ncol(triplet.prior.set) > 1) {
            candidate.pare.set <- list()
            for (jj in seq_len(nrow(triplet.prior.set))) {
              if (length(intersect(parent_of_update, triplet.prior.pare[jj, ])) == 2) {
                aa <- aa + 1
                candidate.pare.set[[aa]] <- triplet.prior.set[jj, ]
              }
            }
            if (length(candidate.pare.set) > 0) {
              score <- numeric()
              for (jj in seq_along(candidate.pare.set)) {
                score[jj] <- candidate.pare.set[[jj]][5]
              }
              bbb <- matrix(nrow = length(candidate.pare.set), ncol = 6)
              for (jjj in seq_along(candidate.pare.set)) {
                bbb[jjj, ] <- candidate.pare.set[[jjj]]
              }

              max.score.candidate <- bbb[bbb[, 5] == max(score), ]
              if (is.numeric(max.score.candidate)) {
                add_parent <- max.score.candidate
              }
              if (is.matrix(max.score.candidate) && nrow(max.score.candidate) > 1) {
                add_parent <- max.score.candidate[sample.int(nrow(max.score.candidate), 1), ]
              }
            }
          }
          prop_incid_matrix <- current_incid_matrix
          prop_ances_matrix <- update_ancestor_matrix(prop_incid_matrix)
          prop_trans_func_matrix <- current_trans_func_matrix
          if (aa > 0) {
            for (i in 1:2) {
              prop_trans_func_matrix[update_order[k], parent_of_update[i]] <- add_parent[4]
            }
          }
          if (prop.prob < prop.ratio || aa == 0) {
            func_order <- sample.int(10, 1)
            for (i in 1:2) {
              prop_trans_func_matrix[update_order[k], parent_of_update[i]] <- func_order
            }
          }
          xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = prop_trans_func_matrix, timeseries = timeseries)
          prop_sample <- xxx[[2]]
          prop_post <- xxx[[1]][length(xxx[[1]])]
          nume <- sum(prop_post)
          deno <- sum(current_post)
          acce_prob <- exp(nume - deno)
          ratio <- runif(1)
          if (acce_prob != 1 && ratio <= acce_prob) {
            n <- n + 1
          }
        }
      }
      if (old == n) { # no update
        iter <- iter + 1
        xxx <- Error_LLH(GeneData = GeneData, SampleSize = SampleSize, num.node = num.node, prior_para = prior_para, penalty = penalty, TRFUM = current_trans_func_matrix, timeseries = timeseries)
        current_sample <- xxx[[2]]
        Sample_Matrix[iter, ] <- current_sample
        all_logpost[iter] <- xxx[[1]][length(xxx[[1]])]
        Incidence_Matrix[[iter]] <- current_incid_matrix
        Ancestor_Matrix[[iter]] <- current_ances_matrix
        Trans_Func_Matrix[[iter]] <- current_trans_func_matrix
      }
      if (old != n) { # when either T or F changes,  recount the number
        iter <- iter + 1
        jump_point[n] <- iter

        logpost[n] <- sum(prop_post)
        Sample_Matrix[iter, ] <- prop_sample
        all_logpost[iter] <- sum(prop_post)
        Incidence_Matrix[[iter]] <- prop_incid_matrix
        Ancestor_Matrix[[iter]] <- prop_ances_matrix
        Trans_Func_Matrix[[iter]] <- prop_trans_func_matrix
      }
    } # end of updating
  } # end of num_update
  # thin to one sample per completed outer iteration (a full Gibbs sweep)
  keep_iter <- seq(floor(burn_in * num_update) + 1L, num_update)
  post_samples <- Trans_Func_Matrix[num.node * keep_iter + 1L]
  # calculate the posterior probability of each edge
  post_edge_prob <- Reduce(`+`, lapply(post_samples, function(m) (m > 0) * 1)) / length(post_samples)
  rownames(post_edge_prob) <- rownames(GeneData)
  colnames(post_edge_prob) <- rownames(GeneData)
  if (verbose) {
    close(pb)
    strong_edges <- sum(post_edge_prob > 0.5)
    # print a clean summary block
    cat("\n")
    cat("=========================================\n")
    cat("          BBNI Analysis Summary          \n")
    cat("=========================================\n")
    cat(sprintf("Nodes Analyzed:          %d\n", num.node))
    cat(sprintf("Samples Processed:       %d\n", SampleSize))
    cat(sprintf("MCMC Iterations:         %d\n", num_update))
    cat(sprintf("Burn-in ratio:           %.2f\n", burn_in))
    cat(sprintf("Final Log-Posterior:     %.3f\n", all_logpost[length(all_logpost)]))
    cat(sprintf("Strong Edges (P > 0.5):  %d\n", strong_edges))
    cat("=========================================\n")
  }
  return(list(
    networks = Trans_Func_Matrix,
    log_posterior = all_logpost,
    post_edge_prob = post_edge_prob,
    burn_in = burn_in
  ))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualize.R
\name{plot_network}
\alias{plot_network}
\title{Plot a Single Boolean Network}
\usage{
plot_network(trans_matrix, node_names = NULL, ...)
}
\arguments{
\item{trans_matrix}{A square matrix representing the network topology and Boolean rules (e.g., the output of \code{GenerateNetwork()}). Elements greater than 0 indicate a directed edge.}

\item{node_names}{Character vector. Optional names for the nodes. Defaults to "N1", "N2", etc.}

\item{...}{Additional graphical parameters passed to \code{igraph::plot.igraph()}.}
}
\value{
An invisible \code{igraph} object.
}
\description{
Visualizes a single directed network topology. For instance, a synthetic "true" network
generated by \code{GenerateNetwork()}.
}
\examples{
# 1. Generate synthetic 5-node network topology
set.seed(123)
true_network <- GenerateNetwork(num.node = 5)

# 2. Plot the generated network
plot_network(true_network)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate.R
\name{GenerateSample}
\alias{GenerateSample}
\title{Simulate Boolean Network Observation Dataset}
\usage{
GenerateSample(
  trans_matrix,
  SampleSize = 50,
  num.node = nrow(trans_matrix),
  para = rep(0.5, nrow(trans_matrix)),
  error = matrix(0, nrow = nrow(trans_matrix), ncol = SampleSize),
  timeseries = TRUE
)
}
\arguments{
\item{trans_matrix}{A square matrix combining the network topology \eqn{T}{T} and integer-coded Boolean logic functions \eqn{F}{F} assigned to each directed edge.}

\item{SampleSize}{An integer representing the total number of samples or time points to simulate. Defaults to 50 if not specified.}

\item{num.node}{An integer representing the total number of network nodes. Defaults to \code{nrow(trans_matrix)} if not specified.}

\item{para}{A numeric vector of baseline success probabilities (\eqn{\theta_i}{\theta_i}) used to generate the expression states of root nodes via independent Bernoulli trials. Defaults to a \code{rep(0.5, nrow(trans_matrix))} if not specified.}

\item{error}{A pre-generated binary noise matrix applied to occasionally flip Boolean outputs, injecting natural noise. Defaults to a zero matrix of size \code{nrow(trans_matrix) x SampleSize} if not specified (no noise).}

\item{timeseries}{Logical. If TRUE, simulates time-lagged data where child state at time \eqn{t}{t} is determined by parent stated at \eqn{t-1}{t-1}. If FALSE, simulates independent samples where child and parent states are resolved simultaneously. Defaults to TRUE.}
}
\value{
A simulated binary gene expression matrix \eqn{G}{G}, where rows represent individual genes/nodes and columns represent samples.
}
\description{
Simulates a synthetic binary observation dataset (\eqn{G}{G}) based on a given
directed acyclic graph topology with Boolean rules. It starts by running
independent Bernoulli trials on root nodes. The remaining non-root nodes
are calculated based on their assigned Boolean logic functions and parent states.
A pre-generated binary noise matrix is applied via a bitwise XOR operation
to occasionally flip the Boolean outputs, injecting natural biological
noise expected by the model.
}
\examples{
# 1. Generate a 5-node network
set.seed(123)
num_nodes <- 5
sample_size <- 10
true_network <- GenerateNetwork(num.node = num_nodes)

# 2. Set baseline probabilities and simulate zero-noise error matrix
root_probs <- rep(0.5, num_nodes)
error_matrix <- matrix(0, nrow = num_nodes, ncol = sample_size)

# 3. Generate cross-sectional (independent mode) synthetic data
dummy_data <- GenerateSample(
  trans_matrix = true_network,
  SampleSize = sample_size,
  para = root_probs,
  error = error_matrix,
  timeseries = FALSE
)
print(dummy_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualize.R
\name{plot_trace}
\alias{plot_trace}
\title{Plot MCMC Trace for BBNI}
\usage{
plot_trace(results, every = 1)
}
\arguments{
\item{results}{The list returned by \code{run_bbni()}, containing \code{networks} and \code{log_posterior}.}

\item{every}{An integer specifying the thinning interval (sampling frequency) for plotting. Default is 1, which plots all log-posterior values. Values greater than 1 plot every \code{every}-th iteration.}
}
\value{
A base R trace plot.
}
\description{
Generates a trace plot of the log-posterior values over iterations of the MCMC
to visually assess convergence and stability of the executed Markov chain.
Burn-in line is graphed to show when the data started to be utilized for
edge-probability calculations.
}
\examples{
# 1. Generate synthetic network and time-series data
set.seed(123)
true_network <- GenerateNetwork(num.node = 5)
dummy_data <- GenerateSample(true_network, SampleSize = 15)

# 2. Run BBNI sampler
prior_para <- matrix(3, nrow = 6, ncol = 2)
prior_para[6, 1] <- 2
prior_para[6, 2] <- 100

results <- run_bbni(dummy_data, prior_para = prior_para, num_update = 100)

# 3. Visualize MCMC results
plot_trace(results)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bbni.r
\name{run_bbni}
\alias{run_bbni}
\title{Execute Metropolis-within-Gibbs MCMC Sampler for Boolean Networks}
\usage{
run_bbni(
  GeneData,
  num.node = nrow(GeneData),
  SampleSize = ncol(GeneData),
  prior_para = NULL,
  num_update = 4000,
  penalty = 0.1,
  prop.ratio = 0.1,
  verbose = FALSE,
  timeseries = TRUE,
  burn_in = 0.7
)
}
\arguments{
\item{GeneData}{A binary empirical observation matrix (\eqn{G}{G}), where rows represent individual network nodes (genes) and columns represent independent samples or sequential time points.}

\item{num.node}{An integer representing the total number of network nodes. Defaults to \code{nrow(GeneData)} if not specified.)}

\item{SampleSize}{An integer representing the total number of time points or independent samples in the dataset. Defaults to \code{ncol(GeneData)} if not specified.}

\item{prior_para}{A matrix (with dimensions \verb{(num.node + 1) x 2}) of Beta prior hyperparameters \eqn{\alpha}{\alpha} and \eqn{\beta}{\beta} for root node probabilities and the global noise parameter \eqn{e}{e}. Defaults to a flat prior if not specified.}

\item{num_update}{An integer representing the total number of MCMC iterations to perform. Defaults to 4000 if not specified.}

\item{penalty}{Structural-prior hyperparameter in \eqn{(0, 1]}{(0, 1]}. A value of \code{1} corresponds to a uniform prior over valid network topologies; values below \code{1} apply an edge-count penalty that favors sparser networks. Defaults to 0.1 if not specified.}

\item{prop.ratio}{A numeric value between 0 and 1 giving the final probability of using the empirical proposal distribution after the first 10\% of outer iterations. During the first 10\% of outer iterations, the empirical proposal is used with probability 0.9. Defaults to 0.1.}

\item{verbose}{Logical. If TRUE, prints verbose MCMC iteration progress to the console. Default is FALSE.}

\item{timeseries}{Logical. If TRUE, the algorithm assumes a time-series dataset. If FALSE, the algorithm assumes independent samples. Default is TRUE.}

\item{burn_in}{A numeric value between 0 and 1 representing the proportion of initial MCMC samples to discard as burn-in. Defaults to 0.7 if not specified.}
}
\value{
A list containing the full trajectory of the MCMC chain. Specifically, \code{networks} (a list of sampled transition function matrices) and \code{log_posterior} (a numeric vector of log-posterior scores for each iteration). These represent samples drawn from the marginal posterior distribution \eqn{P(T,F|G)}{P(T,F|G)} used for Bayesian model averaging. Additionally, the \code{post_edge_prob} (matrix of marginal posterior edge probabilities) and \code{burn_in} ratio are returned in the list.
}
\description{
Executes a Metropolis-within-Gibbs Markov chain Monte Carlo (MCMC) algorithm to sample
from the joint posterior distribution of Directed Acyclic Graph (DAG) topologies (\eqn{T}{T}) and
Boolean logic transition functions (\eqn{F}{F}). The algorithm iterates through individual
network nodes and proposes parent set mutations (edge additions, removals, or swaps)
paired with transition function reassignments to one of 14 candidate Boolean rules (stored as codes 1-12 in the transition matrix).
Proposed states transitions are strictly verified to follow the DAG constraint and
evaluated with a Metropolis-Hastings acceptance threshold using log-posterior values.
}
\details{
Posterior edge probabilities (\code{post_edge_prob}) are computed from one
thinned sample per outer iteration (a full Gibbs sweep) after discarding
\code{burn_in}, as designed in the original method paper.
}
\examples{
# 1. Define network parameters
set.seed(235)
num_nodes <- 8
sample_size <- 50

# 2. Generate true network and simulate data
true_network <- GenerateNetwork(num.node = num_nodes)

# Set up Beta priors for root-node probabilities and the noise rate
prior_para <- matrix(3, nrow = num_nodes + 1, ncol = 2)
prior_para[num_nodes + 1, 1] <- 2
prior_para[num_nodes + 1, 2] <- 100

# Simulate parameters
para <- numeric(num_nodes + 1)
for (i in 1:(num_nodes + 1)) {
  para[i] <- stats::rbeta(1, prior_para[i, 1], prior_para[i, 2])
}
para[num_nodes + 1] <- 0.1 # Fixed noise rate for simulation

error_matrix <- matrix(stats::rbinom(num_nodes * sample_size, 1, para[num_nodes + 1]),
  nrow = num_nodes, ncol = sample_size
)

dummy_data <- GenerateSample(
  trans_matrix = true_network,
  SampleSize = sample_size,
  para = para,
  error = error_matrix
)

# 3. Run the MCMC sampler (silently)
mcmc_results <- run_bbni(
  GeneData = dummy_data,
  prior_para = prior_para,
  num_update = 80, # Scaled down for example speed
  prop.ratio = 0.1
)

# 4. Visualize results
plot_bbni(mcmc_results, true_network = true_network, threshold = 0.5)
plot_trace(mcmc_results)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BBNI-package.R
\docType{package}
\name{BBNI-package}
\alias{BBNI}
\alias{BBNI-package}
\title{BBNI: Bayesian Inference of Boolean Genetic Networks}
\description{
Implements a fully Bayesian Markov chain Monte Carlo (MCMC) approach for inferring the topology and Boolean logic transition functions of gene regulatory networks from noisy, binary time-series expression data. Network structure and Boolean rules are sampled jointly from their posterior distribution, providing principled uncertainty quantification rather than a single point estimate. Method described in Han et al. (2014) \doi{10.1371/journal.pone.0115806}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://anson-li8.github.io/BBNI/}
  \item \url{https://github.com/anson-li8/BBNI}
  \item Report bugs at \url{https://github.com/anson-li8/BBNI/issues}
}

}
\author{
\strong{Maintainer}: Anson Li \email{liyuanrui618@gmail.com}

Authors:
\itemize{
  \item Anson Li \email{liyuanrui618@gmail.com}
  \item Shengtong Han \email{shengtong.han@marquette.edu}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate.R
\name{GenerateNetwork}
\alias{GenerateNetwork}
\title{Generate Initial Network Topology}
\usage{
GenerateNetwork(num.node)
}
\arguments{
\item{num.node}{An integer representing the total number of genes/nodes in the network.}
}
\value{
A square transition function matrix combining the initial DAG topology with randomly assigned Boolean logic functions. Elements with a value of 0 indicate no directed edge, while positive integers indicate the presence of an edge and specify the defining Boolean function type (codes 1-12: 1-10 for two-input functions, 11-12 for one-input identity/negation functions).
}
\description{
Randomly generates a valid directed acyclic graph (DAG) topology
\eqn{T}{T} and assigns a corresponding Boolean transition function \eqn{F}{F} to each node.
The algorithm samples parent set configurations, keeping the constraint
that the maximum in-degree for any node is 2, and further ensures
the resulting structure does not contain directed cyclic loops.
}
\examples{
# Generate a true network topology and Boolean rules for 5 nodes
set.seed(123)
true_network <- GenerateNetwork(num.node = 5)

# Graph the network with built-in exported function plot_network
plot_network(true_network)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualize.R
\name{plot_bbni}
\alias{plot_bbni}
\title{Plot the Inferred Bayesian Boolean Network}
\usage{
plot_bbni(
  results,
  threshold = 0.5,
  node_names = NULL,
  true_network = NULL,
  ...
)
}
\arguments{
\item{results}{The list returned by \code{run_bbni()}, containing \code{networks} and \code{log_posterior}.}

\item{threshold}{Numeric. The minimum posterior probability required to draw an edge. Defaults to 0.5.}

\item{node_names}{Character vector. Optional names for the nodes. Defaults to "N1", "N2", etc.}

\item{true_network}{Optional square matrix representing the true network topology. If provided, edges will be color-coded to indicate true positives (along with displaying wrong function inferences), false positives, and false negatives. Purely for simulation purposes.}

\item{...}{Additional graphical parameters passed to \code{igraph::plot.igraph()}.}
}
\value{
An invisible \code{igraph} object.
}
\description{
Visualizes the causal network structure inferred by the BBNI MCMC sampler.
This function takes the marginal posterior probability of each directed edge
and plots the network using the \code{igraph} package. Edges with posterior
probabilities below the specified threshold are omitted from the plot.
}
\examples{
# 1. Generate synthetic network and time-series data
set.seed(123)
true_network <- GenerateNetwork(num.node = 5)
dummy_data <- GenerateSample(true_network, SampleSize = 15)

# 2. Run BBNI sampler
prior_para <- matrix(3, nrow = 6, ncol = 2)
prior_para[6, 1] <- 2
prior_para[6, 2] <- 100

results <- run_bbni(dummy_data, prior_para = prior_para, num_update = 100)

# 3. Plot inferred network
plot_bbni(results, true_network = true_network, threshold = 0.5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{yeast_data}
\alias{yeast_data}
\title{Yeast Cell-Cycle Gene Expression Data}
\format{
A numeric matrix with 14 rows (representing individual genes/nodes)
and 385 columns (representing independent samples or sequential time points).
}
\source{
Han, S., Wong, R. K. W., Lee, T. C. M., Shen, L., Li, S.-Y. R., & Fan, X. (2014).
A Full Bayesian Approach for Boolean Genetic Network Inference. \emph{PLOS ONE}, 9(12), e115806.
\doi{10.1371/journal.pone.0115806}
}
\usage{
data(yeast_data)
}
\description{
A binary empirical dataset with the gene expression states of 14 yeast
cell-cycle genes across 385 complete pooled experimental conditions. The
data were filtered to the 14 genes used by Han et al. (2014), columns with
missing values were removed, and each gene was binarized by its row mean.
}
\keyword{datasets}
